local Events = require ( "Events" )
local ModelEvents = Events.E_MODEL
local GameEvents = Events.E_GAME
local NetworkEvents = Events.E_NETWORK

local GameView = require ( "View.GameView" )
local GameModel = require ( "Model.GameModel" )
local LocalPlayerCtrl = require ( "Controller.PlayerCtrl.LocalPlayerCtrl" )
local AIPlayerCtrl = require ( "Controller.PlayerCtrl.AIPlayerCtrl" )
local NetPlayerCtrl = require ( "Controller.PlayerCtrl.NetPlayerCtrl" )
local Network = require ( "Network" )
local SoundManager = require ( "SoundManager" )
local AppCtrl = core.class ( Object )

local function newTimer ( span, callback, loop )
    local timer = MOAITimer.new ()
    timer:setSpan ( span )
    if loop then
        timer:setMode ( MOAITimer.LOOP )
        timer:setListener ( MOAITimer.EVENT_TIMER_LOOP, callback )
    else
        timer:setMode ( MOAITimer.NORMAL )
        timer:setListener ( MOAITimer.EVENT_TIMER_END_SPAN, callback )
    end
    timer:start ()
    return timer
end

function AppCtrl:ctor ()
    local fsm = FSM.new ( self )
    self.fsm = fsm
    
    self._gameMode = 301
    
    MOAIGfxDevice.setListener ( MOAIGfxDevice.EVENT_RESIZE, function ( width, height ) self.view:resize ( width, height ) end )
    
    app.queue:subscribe ( fsm, FSM.handleEvent, E_SYSTEM )
    app.queue:subscribe ( fsm, FSM.handleEvent, ModelEvents )
    app.queue:subscribe ( fsm, FSM.handleEvent, GameEvents )
    app.queue:subscribe ( fsm, FSM.handleEvent, NetworkEvents )

    fsm:declare
    {
        {
            enter = function ( self )
            end;

            leave = function ( self )
                Network.open ()
                SoundManager.init ()
                self.session = { isHost = false, id = MOAIEnvironment.generateGUID ()}
                if Network.isAvailable () then
                    --log ( "\n>>>> Network is available\n" )
                else
                    --log ( "\n>>>> Network is not available\n" )
                end
            end;

            [ E_SYSTEM.START ] = function ( self, sender, event, args )
                    return "SPLASH"
                end;
        };
        
        SPLASH = 
        {
            enter = function ( self )
                local view = GameView.new ()
                view:showSplashScreen ()
                self.view = view
                self._remoteReady = false
            end;
            
            [ NetworkEvents.READY ] = function ( self, sender, event, args )
                self.session.isHost = self.session.id < args.id
                self._remoteReady = true
            end;
            
            [ E_SYSTEM.SESSION_STARTED ] = function ( self, sender, event, args )
                --log ( "\n>>>> Network session started\n" )
                self.view:showHint ( "Waiting for friend" )
            end;
           
            [ E_SYSTEM.LEAVE_SPLASH ] = function ( self, sender, event, args )
                if self._remoteReady then
                    --log ( "\n>>>> Multiplayer mode activated\n" )
                    self.isMultiplayer = true
                    Network.sendMessage ( Network.MT_READY, { id = self.session.id })
                    return "INIT"
                else
                    if Network.isAvailable () then
                        return "CONNECTION"
                    else
                        --log ( "\n>>>> Singleplayer mode activated\n" )
                        self.isMultiplayer = false
                        return "INIT"
                    end
                end
            end;
        };
        
        CONNECTION =
        {
            enter = function ( self )
                local view = self.view
                view:setAllowedGestures ( GameView.GESTURES.TAP )
               
                local waitForReply = function ( self )
                    view:hideTextPopup ()
                    Network.sendMessage ( Network.MT_READY, { id = self.session.id })
                    view:setAllowedGestures ( GameView.GESTURES.NONE )
                    self._connectionTimer:stop ()
                    self._connectionTimer:setSpan ( 20.0 )
                    self._connectionTimer:start ()
                end
                
                self._connectionTimer = newTimer ( 0.5,
                        function ()
                            self._connectionTimer = nil
                            self._remoteReady = false
                            app.queue:postEvent ( nil, E_SYSTEM.LEAVE_CONNECTION )
                        end
                    )
                
                if Network.isConnected () then
                    waitForReply ( self )
                else
                    self._pingTimer = newTimer ( 0.1, 
                        function ()
                            if Network.isConnected () then
                                waitForReply ( self )
                                self._pingTimer:stop ()
                                self._pingTimer = nil
                            end
                        end,
                        true
                    )
                end
            end;
            
            leave = function ( self )
                if self._pingTimer then
                    self._pingTimer:stop ()
                    self._pingTimer = nil
                end
                self.view:hideTextPopup ()
                self._remoteReady = nil
            end;
            
            [ GameEvents.DEFAULT_TAP ] = function ( self, sender, event, args )
                if self._connectionTimer then
                    self._connectionTimer:stop ()
                    self._connectionTimer = nil
                    self._remoteReady = false
                    app.queue:postEvent ( nil, E_SYSTEM.LEAVE_CONNECTION )
                end
            end;
            
            [ E_SYSTEM.LEAVE_CONNECTION ] = function ( self, sender, event, args )
                if self._remoteReady then
                    self.isMultiplayer = true
                    --log ( "\n>>>> Multiplayer mode activated\n" )
                else
                    --log ( "\n>>>> Singleplayer mode activated\n" )
                end
                return "INIT"
            end;
            
            [ E_SYSTEM.SESSION_STARTED ] = function ( self, sender, event, args )
                --log ( "\n>>>> Network session started\n" )
                self.view:showHint ( "Waiting for friend" )
            end;
            
            [ NetworkEvents.READY ] = function ( self, sender, event, args )
                if self._connectionTimer then
                    self._connectionTimer:stop ()
                    self._connectionTimer = nil
                    self._remoteReady = true
                    app.queue:postEvent ( nil, E_SYSTEM.LEAVE_CONNECTION )
                    self.session.isHost = self.session.id < args.id
                end
            end;
        };
        
        INIT = 
        {
            enter = function ( self )
                local model = GameModel.new ()
                local view = self.view
                model:setup ()
                
                self.view:hideHint ()
                view:setModel ( model )
                
                local p1, p2
                local i1, i2 = GameModel.PLAYER_1, GameModel.PLAYER_2
                if self.isMultiplayer then
                    if not self.session.isHost then
                        i1, i2 = i2, i1
                    end
                    p1 = LocalPlayerCtrl.new ( model, view, i1 )
                    p2 = NetPlayerCtrl.new ( model, view, i2 )
                    p1:setMultiplayer ( true )
                    p2:setMultiplayer ( true )
                else
                    p1 = LocalPlayerCtrl.new ( model, view, i1 )
                    p2 = AIPlayerCtrl.new ( model, view, i2 )
                end
                
                local players = { [ i1 ] = p1, [ i2 ] = p2 }
                
                self.model = model
                self.players = players
                
                view:showGameScreen ()
                return "GAME_SETUP"
            end;
        };
        
        GAME_SETUP = 
        {
            enter = function ( self )
                self.view:displayTextPopup ( "301 GAME" )
                self._gameSetupTimer = newTimer ( 1.0,
                        function ()
                            self._gameSetupTimer = nil
                            app.queue:postEvent ( nil, GameEvents.DEFAULT_TAP )
                        end
                    )
                if self.isMultiplayer then
                    if self.session.isHost then
                        self._firstPlayer = math.random ( GameModel.PLAYER_2 )
                        Network.sendMessage ( Network.MT_START, { firstPlayer = self._firstPlayer })
                    end
                else
                    self._firstPlayer = GameModel.PLAYER_1
                end
            end;
            
            leave = function ( self )
                if self._gameSetupTimer then
                    self._gameSetupTimer:stop ()
                    self._gameSetupTimer = nil
                end
                self.view:hideTextPopup ()
            end;
            
            [ NetworkEvents.START ] = function ( self, sender, event, args )
                self._firstPlayer = args.firstPlayer
                if not self._gameSetupTimer then
                    app.queue:postEvent ( nil, GameEvents.DEFAULT_TAP )
                end
            end;
            
            [ GameEvents.DEFAULT_TAP ] = function ( self, sender, event, args )
                if self._firstPlayer then
                    return "GAME"
                end
            end;
        };
        
        GAME = 
        {
            enter = function ( self )
                self.model:start ( self._gameMode, self._firstPlayer )
            end;
            
            [ ModelEvents.STATE_CHANGED ] = function ( self, sender, event, args )
                local model = sender
                if args.nextState == model.STATE_ROUND_BEG then
                    model:startRound ()
                end
                if args.nextState == model.STATE_TURN_BEG then
                    model:startTurn ()
                    self._activePlayerIndex = self.model:activePlayer (): index ()
                    self.players [ self._activePlayerIndex ]: activate ()
                end
                if args.nextState == model.STATE_TURN_END then
                    self.players [ self._activePlayerIndex ]: deactivate ()
                    model:nextTurn ()
                end
                if args.nextState == model.STATE_ROUND_END then
                    self.players [ self._activePlayerIndex ]: deactivate ()
                    return "GAME_SETUP"
                end
                if args.nextState == model.STATE_DONE then
                    self.players [ self._activePlayerIndex ]: deactivate ()
                    self.view:clearBoard ()
                    return "GAME_OVER"
                end
            end;
            
            [ GameEvents.QUIT_PROMPT ] = function ( self, sender, event, args )
                self.view:displayConfirmPopup ( "QUIT", { text = "YES", event = GameEvents.QUIT }, { text = "NO", event = GameEvents.BACK })
            end;
            
            [ GameEvents.INFO ] = function ( self, sender, event, args )
                self.view:showInfo ()
            end;
            
            [ GameEvents.QUIT ] = function ( self, sender, event, args )
                return "FINAL"
            end;
            
            [ GameEvents.BACK ] = function ( self, sender, event, args )
                self.view:hideConfirmPopup ()
            end;
        };
        
        GAME_OVER = 
        {
            enter = function ( self )
                return "FINAL"
            end;
            
            leave = function ( self )
            end;
            
            [ GameEvents.DEFAULT_TAP ] = function ( self, sender, event, args )
                return "FINAL"
            end;
            
            [ GameEvents.QUIT ] = function ( self, sender, event, args )
                return "FINAL"
            end;
        };
        
        FINAL = 
        {
            enter = function ( self )
                Network.close ()
                if MOAISim.exit then
                    MOAISim.exit ()
                else
                    os.exit ()
                end
            end;
        };
    }
end

return AppCtrl
