local core = require ( "core" )

local GameView = require ( "View.GameView" )
local GameModel = require ( "Model.GameModel" )
local BoardModel = require ( "Model.BoardModel" )

local SoundManager = require ( "SoundManager" )

local Events = require ( "Events" )
local GameEvents = Events.E_GAME

local PlayerCtrl = core.class ( Object )

function PlayerCtrl.isLocalPlayer ()
    return false
end

function PlayerCtrl.name ()
    return "PLAYER"
end

function PlayerCtrl.turnLabel ()
    return "TURN"
end

function PlayerCtrl:ctor ( model, view, index )
    self._model = model
    self._view = view
    self._index = index
    model:players () [ index ]:setName ( self.name ())
end

function PlayerCtrl:activate ()
    self.isActive = true
    local view = self._view
    view:clearBoard ()
    view.sigDartsThrow:connect ( self, self.onDartsThrow )
    view.sigDartsHit:connect ( self, self.onDartsHit )
    view:displayTextPopup ( self.turnLabel ())
    if not self.isLocalPlayer () then
        view:displayTurnLabel ( self._index, self.turnLabel ())
    end
    self._timer = self:performAfterDelay ( 1.0, self.hmActivated )
end

function PlayerCtrl:deactivate ()
    self._view.sigDartsHit:disconnect ( self, self.onDartsHit )
    self._view.sigDartsThrow:disconnect ( self, self.onDartsThrow )
    self._view:hideTurnLabel ()
    self.isActive = false
end

function PlayerCtrl:performAfterDelay ( delay, func, ... )
    local timer = MOAITimer.new ()
    local args = {...}
    timer:setSpan ( delay )
    timer:setMode ( MOAITimer.NORMAL )
    timer:setListener ( MOAITimer.EVENT_TIMER_END_SPAN,
        function ()
            func ( self, unpack ( args ))
        end
    )
    timer:start ()
    return timer
end

function PlayerCtrl:_hideHint ()
    if self._hintTimer then
        self._hintTimer:stop ()
        self._hintTimer = nil
        self._view:showHint ( "" )
    end
end

function PlayerCtrl:_hideHighlight ()
    if self._highlightTimer then
        self._highlightTimer:stop ()
        self._highlightTimer = nil
        self._view:unhiglightSector ( "" )
    end
end

function PlayerCtrl:onDartsThrow ( x, y, vx, vy )
    local model = self._model
    local x, y = self._view:finalModelCoordinates ( x, y, vx, vy )
    local value = model:board ():getTargetedValue ( x, y )

    self:performAfterDelay ( 0.2, 
        function ( self )
            SoundManager.play ( "hit.ogg" )
        end
    )

    if ( model:activePlayer ():currentScore () - value < 0 ) then
        self:performAfterDelay ( 0.7, 
            function ( self )
                SoundManager.play ( "bust.ogg" )
            end
        )
    end
end

function PlayerCtrl:onDartsHit ( x, y )
    local model = self._model
    local view = self._view
    
    local hitResult, value, circle, sValue = model:processHit ( x, y )
    
    view:displayScore ( value, sValue, circle )
    
    self:_hideHighlight ()
    if circle < BoardModel.R_FULL then
        local sector = model:board ():getSectorByValue ( sValue )
        view:higlightSector ( sector, circle )
        self._highlightTimer = self:performAfterDelay ( 1.0, self._hideHighlight )
    end
    
    local hint = nil
    if circle == BoardModel.R_DOUBLE_RING then
        hint = "DOUBLE " .. sValue
    elseif circle == BoardModel.R_TRIPLE_RING then
        hint = "TRIPLE " .. sValue
    elseif circle == BoardModel.R_BULLS_EYE then
        hint = "BULLSEYE"
    end
    if hint then
        self:_hideHint ()
        view:showHint ( hint )
        self._hintTimer = self:performAfterDelay ( 1.2, self._hideHint )
    end
    
    if hitResult == GameModel.HIT_BUST then
        local info = ""
        if self.isLocalPlayer () then
            info = "Reduce score exactly to zero"
        end
        view:displayTextPopup ( "BUST", info )
        self._timer = self:performAfterDelay ( 2.0, self.hmEndTurn )
    elseif hitResult == GameModel.HIT_TURN then
        self._timer = self:performAfterDelay ( 1.0, self.hmEndTurn )
    elseif hitResult == GameModel.HIT_WIN then
        local isLocal = self.isLocalPlayer ()
        local activePlayer = model:activePlayer ():index ()
        self._timer = self:performAfterDelay ( 1.0,
            function ( self )
                if isLocal then
                    view:displayTextPopup ( "YOU WIN", "Tap to exit" )
                    SoundManager.play ( "win.ogg" )
                else
                    view:displayTextPopup ( "YOU LOSE", "Tap to exit", activePlayer % 2 + 1 )
                    SoundManager.play ( "lose.ogg" )
                end
            end
        )
    end
    return hitResult, value, circle, sValue
end

function PlayerCtrl:hmActivated ()
    local model = self._model
    local view = self._view
    
    if self._timer then
        view:hideTextPopup ()
        self._timer:stop ()
        self._timer = nil
    end
end

function PlayerCtrl:hmEndTurn ()
    local model = self._model
    local view = self._view
    
    if self._timer then
        view:hideTextPopup ()
        self._timer:stop ()
        self._timer = nil
    end
    model:finishTurn ()
end

function PlayerCtrl:hmEndRound ()
    local model = self._model
    local view = self._view
    
    if self._timer then
        self._timer:stop ()
        self._timer = nil
    end
    view:hideTextPopup ( true )
    model:finishRound ()
end

function PlayerCtrl:setMultiplayer ( multiplayer )
    self._isMultiplayer = multiplayer
end

function PlayerCtrl:isMultiplayer ()
    return self._isMultiplayer
end

return PlayerCtrl
