local core = require ( "core" )

local GameView = require ( "View.GameView" )
local GameModel = require ( "Model.GameModel" )
local BoardModel = require ( "Model.BoardModel" )

local SoundManager = require ( "SoundManager" )

local Events = require ( "Events" )
local GameEvents = Events.E_GAME
local AnimEvents = Events.E_ANIMATION

local PlayerCtrl = require ( "Controller.PlayerCtrl" )

local LocalPlayerCtrl = core.class ( PlayerCtrl )
{
    HELP_RADIUS = 15,
}

function LocalPlayerCtrl.isLocalPlayer ()
    return true
end

function LocalPlayerCtrl.name ()
    return "YOU"
end

function LocalPlayerCtrl.turnLabel ()
    return "YOUR TURN"
end

function LocalPlayerCtrl:ctor ( model, view , index)
    self.super.ctor ( self, model, view, index )
    self._firstThrow = true
    local board = model:board ()
    self._doubleMiddleRadius = ( board:getCircle ( BoardModel.R_DOUBLE_RING ).r + board:getCircle ( BoardModel.R_OUTER_RING ).r ) * 0.5
    self._swipeHintTimer = self:performAfterDelay ( 10.0, self.showSwipeHint )
    self._swipeHintTimer:stop ()
    
    self._canHint = true
    app.queue:subscribe ( self, LocalPlayerCtrl.onPause, GameEvents.INFO )
    app.queue:subscribe ( self, LocalPlayerCtrl.onResume, GameEvents.HIDE_INFO )
    app.queue:subscribe ( self, LocalPlayerCtrl.onPause, GameEvents.QUIT_PROMPT )
    app.queue:subscribe ( self, LocalPlayerCtrl.onResume, GameEvents.BACK )
end

function LocalPlayerCtrl:activate ()
    self.super.activate ( self )
    
    local model = self._model
    local view = self._view
    
    app.queue:subscribe ( self, self.hmActivated, GameEvents.DEFAULT_TAP )
    view:setAllowedGestures ( GameView.GESTURES.TAP )
    
    view.sigGesBegan:connect ( self, self.onGesBegan )
    view.sigTapEnded:connect ( self, self.onTapEnded )
    view.sigPanMoved:connect ( self, self.onPanMoved )
    view.sigPanEnded:connect ( self, self.onPanEnded )
    
    app.queue:subscribe ( self, LocalPlayerCtrl.onSwipeHintEnded, AnimEvents.SWIPE_HINT )
    local player = model:players() [ self._index ]
    app.queue:subscribe ( self, LocalPlayerCtrl.onDartsReady, AnimEvents.DARTS_READY, player )
    app.queue:subscribe ( self, LocalPlayerCtrl.onDartsDropped, AnimEvents.DARTS_DROPPED, player )

    SoundManager.play ( "start.ogg" )
    if self._canHint then
        self._swipeHintTimer:start ()
    end
end

function LocalPlayerCtrl:onPause ()
    self._canHint = false
    self._swipeHintTimer:stop ()
    self:hideSwipeHint ()
end

function LocalPlayerCtrl:onResume ()
    self._canHint = true
    self._swipeHintTimer:start ()
end

function LocalPlayerCtrl:showSwipeHint ()
    if not self._showingHint then
        self._showingHint = true
        self._swipeHintTimer:stop ()
        self._view:showSwipeHint ()
    end
end

function LocalPlayerCtrl:onSwipeHintEnded ()
    if self._showingHint then
        self._view:showSwipeHint ()
    end
end

function LocalPlayerCtrl:hideSwipeHint ()
    if self._showingHint then
        self._view:hideSwipeHint ()
        self._showingHint = false
    end
end

function LocalPlayerCtrl:onGesBegan ( x, y, canCreate )
    self:hideSwipeHint ()
    self._swipeHintTimer:stop ()
    self._dartsStart = false
    if self._canThrow and canCreate and self._view:canGrabDart ( x, y ) then
        self._dartsStart = true
        self._view:dartsAiming ( x, y )
        self._view:highlightDarts ()
    end
end

function LocalPlayerCtrl:onTapEnded ( x, y )
    if self._canHint then
        self._swipeHintTimer:start ()    
    end
    if self._dartsStart then
        self._dartsStart = false
        self._view:endDartsAiming ( 0, 0 )
        self._view:unhighlightDarts ()
    end
end

function LocalPlayerCtrl:onPanMoved ( x, y )
    if self._dartsStart then
        self._view:dartsAiming ( x, y )
    end
end

function LocalPlayerCtrl:onPanEnded ( x, y, vx, vy )
    if self._canHint then
        self._swipeHintTimer:start ()
    end
    
    if self._dartsStart then
        local view = self._view
        local model = self._model
        
        local dx, dy = 0, 0
        
        if not self._isMultiplayer then
            local board = model:board ()
            local x, y = view:finalModelCoordinates ( x, y, vx, vy )
            local r, f = BoardModel.boardCoordinates ( x, y )
            local dist = math.abs ( r - self._doubleMiddleRadius )
            if dist < LocalPlayerCtrl.HELP_RADIUS then
                local doubleX, doubleY = BoardModel.modelCoordinates ( self._doubleMiddleRadius, f )
                local doubleValue = board:getTargetedValue ( doubleX, doubleY )
                if model:activePlayer ():currentScore () - doubleValue == 0 then
                    dx, dy = ( doubleX - x ) / GameView.VELOCITY_SCALE_FACTOR , ( doubleY - y ) / GameView.VELOCITY_SCALE_FACTOR
                end
            end
        end
        
        local vx, vy = vx + dx, vy + dy
        view:endDartsAiming ( vx, vy )
        view:unhighlightDarts ()
    end
end

function LocalPlayerCtrl:onDartsThrow ( x, y, vx, vy )
    self.super.onDartsThrow ( self, x, y, vx, vy )
    local model = self._model
    local view = self._view
    
    view:setAllowedGestures ( GameView.GESTURES.TAP )
    self._canThrow = false
    if self._hint then
        view:hideHint ()
        self._hint = nil
    end
end

function LocalPlayerCtrl:onDartsReady ()
    local view = self._view
    if self._hint then
        view:setHintAlpha ( 0.0 )
    end
end

function LocalPlayerCtrl:onDartsDropped ()
    local view = self._view
    if self._hint then
        view:setHintAlpha ( 1.0 )
    end
end

function LocalPlayerCtrl:onDartsHit ( x, y )
    local model = self._model
    local view = self._view
    
    local hitResult, value = self.super.onDartsHit ( self, x, y )
    if hitResult == GameModel.HIT_BUST then
        app.queue:subscribe ( self, self.hmEndTurn, GameEvents.DEFAULT_TAP )
        view:setAllowedGestures ( GameView.GESTURES.TAP )
    elseif hitResult == GameModel.HIT_TURN then
        app.queue:subscribe ( self, self.hmEndTurn, GameEvents.DEFAULT_TAP )
        view:setAllowedGestures ( GameView.GESTURES.TAP )
    elseif hitResult == GameModel.HIT_WIN then
        self._swipeHintTimer:stop ()
        app.queue:subscribe ( self, self.hmEndRound, GameEvents.DEFAULT_TAP )
        view:setAllowedGestures ( GameView.GESTURES.TAP )
    else
        view:setAllowedGestures ( GameView.GESTURES.TAP_PAN )
        self._canThrow = true
    end
end

-- function LocalPlayerCtrl:showOutHint ()
--     local player = self._model:activePlayer ()
--     local score = player:currentScore ()
--     if not self._isMultiplayer and score <= 40  then
--         if score % 2 == 0 then
--             self._view:showHint ( "Hit double " .. tostring ( score / 2 ) .. "\n to win the game")
--             self._hint = true
--         end
--     end
-- end

function LocalPlayerCtrl:hmActivated ()
    app.queue:unsubscribe ( self, self.hmActivated, GameEvents.DEFAULT_TAP )
    self.super.hmActivated ( self )
    local view = self._view
    view:setAllowedGestures ( GameView.GESTURES.TAP_PAN )
    self._canThrow = true
    view:setHintAlpha ( 1.0 )
    if self._firstThrow then
        view:showHint ( "Tap and hold, then swipe\nto throw a dart" )
        self._firstThrow = nil
        self._hint = true
    end
end

function LocalPlayerCtrl:hmEndTurn ()
    app.queue:unsubscribe ( self, self.hmEndTurn, GameEvents.DEFAULT_TAP )
    self.super.hmEndTurn ( self )
end

function LocalPlayerCtrl:hmEndRound ()
    app.queue:unsubscribe ( self, self.hmEndRound, GameEvents.DEFAULT_TAP )
    self.super.hmEndRound ( self )
end

function LocalPlayerCtrl:deactivate ()
    local model = self._model
    local view = self._view

    app.queue:unsubscribe ( self, LocalPlayerCtrl.onDartsDropped, AnimEvents.DARTS_DROPPED )
    app.queue:unsubscribe ( self, LocalPlayerCtrl.onDartsReady, AnimEvents.DARTS_READY )
    app.queue:unsubscribe ( self, LocalPlayerCtrl.onSwipeHintEnded, AnimEvents.SWIPE_HINT )
    
    view.sigGesBegan:disconnect ( self, self.onGesBegan )
    view.sigTapEnded:disconnect ( self, self.onTapEnded )
    view.sigPanMoved:disconnect ( self, self.onPanMoved )
    view.sigPanEnded:disconnect ( self, self.onPanEnded )
    
    view:setAllowedGestures ( GameView.GESTURES.TAP )
    
    self:hideSwipeHint ()
    self._swipeHintTimer:stop ()
    self._canThrow = false
 
    self.super.deactivate ( self )
end

return LocalPlayerCtrl
