require ( "core" )

local GestureRecognizer = core.class ( Object )
{
    GR_GESTURE_NONE = 1,
    GR_GESTURE_BEGAN = 2,
    GR_TAP_ENDED = 3,
    GR_PAN_MOVED = 4,
    GR_PAN_ENDED = 5,
    
    DEFAULT_TAP_RADIUS = 2,
    MAX_DELTA_TIME = 0.5,
}

local TOUCH_UP = MOAITouchSensor.TOUCH_UP
local TOUCH_DOWN = MOAITouchSensor.TOUCH_DOWN
local TOUCH_MOVE = MOAITouchSensor.TOUCH_MOVE
local TOUCH_CANCEL = MOAITouchSensor.TOUCH_CANCEL

function GestureRecognizer._touchDistance ( touch0, touch1 )
    local dx, dy = touch1 [ 1 ] - touch0 [ 1 ], touch1 [ 2 ] - touch0 [ 2 ]
    return math.sqrt ( dx * dx + dy * dy )
end

function GestureRecognizer:ctor ( context )
    local fsm = FSM.new ( self, true )
    self._fsm = fsm
    
    self.tapRadius = GestureRecognizer.DEFAULT_TAP_RADIUS
    
    self._gestureCallbacks = {}
    fsm:declare
    {
        [ GestureRecognizer.GR_GESTURE_NONE ] = 
        {
            [ TOUCH_DOWN ] = function ( self, sender, event, args )
                self._startTouch = args
                self._prevTouch = args
                return self:_switchState( GestureRecognizer.GR_GESTURE_BEGAN, context, args )
            end;
        };
        
        [ GestureRecognizer.GR_GESTURE_BEGAN ] = 
        {
            [ TOUCH_MOVE ] = function ( self, sender, event, args )
                if GestureRecognizer._touchDistance ( self._startTouch, args ) > self.tapRadius then
                    return self:_switchState( GestureRecognizer.GR_PAN_MOVED, context, args )
                end
            end;
            
            [ TOUCH_UP ] = function ( self, sender, event, args )
                return self:_switchState( GestureRecognizer.GR_TAP_ENDED, context, args )
            end;
        };
    
        [ GestureRecognizer.GR_TAP_ENDED ] = 
        {
            enter = function ( self )
                return self:_switchState( GestureRecognizer.GR_GESTURE_NONE, context, args )
            end;
        };
        
        [ GestureRecognizer.GR_PAN_MOVED ] = 
        {
            [ TOUCH_MOVE ] = function ( self, sender, event, args )
                self:_invokeCallback ( GestureRecognizer.GR_PAN_MOVED, context, args )
            end;
            
            [ TOUCH_UP ] = function ( self, sender, event, args )
                return self:_switchState( GestureRecognizer.GR_PAN_ENDED, context, args )
            end;
        };
        
        [ GestureRecognizer.GR_PAN_ENDED ] = 
        {
            enter = function ( self )
                return self:_switchState( GestureRecognizer.GR_GESTURE_NONE, context, args )
            end;
        };
    }
end

function GestureRecognizer:setCallback ( state, callback )
    self._gestureCallbacks [ state ] = callback
end

function GestureRecognizer:_switchState ( nextState, context, touch )
    self:_invokeCallback ( nextState, context, touch )
    return nextState
end

function GestureRecognizer:_invokeCallback ( state, context, touch )
    local callback = self._gestureCallbacks [ state ]
    if callback then
        callback ( context, state, touch )
    end
end

function GestureRecognizer:initInput ()
    if MOAIInputMgr.device.pointer then
        self._touchDown = false
        MOAIInputMgr.device.pointer:setCallback (
            function ( x, y )
                if self._touchDown then
                    self:handleTouch ( TOUCH_MOVE, x, y )
                end
            end
        )
        MOAIInputMgr.device.mouseLeft:setCallback (
            function ( isMouseDown )
                local x, y = MOAIInputMgr.device.pointer:getLoc ()
                self._touchDown = isMouseDown
                self:handleTouch ( isMouseDown and TOUCH_DOWN or TOUCH_UP, x, y )
            end
        )
    else
        MOAIInputMgr.device.touch:setCallback (
            function ( eventType, idx, x, y, tapCount )
                if idx == 0 then
                    self:handleTouch ( eventType, x, y )
                end
            end
        )
    end
    self._prevTouch = { 0, 0, 0 }
    self:enableInput ()
end

function GestureRecognizer:enableInput ()
    if not self._inputEnabled then
        self._inputEnabled = true
    end
end

function GestureRecognizer:disableInput ()
    if self._inputEnabled then
        self._inputEnabled = false
    end
end

function GestureRecognizer:handleTouch ( eventType, x, y )
    if self._inputEnabled then
        local eventType = eventType == TOUCH_CANCEL and TOUCH_UP or eventType
        local timestamp = MOAISim:getElapsedTime ()
        local touch = { x, y, timestamp }
        self._fsm:handleEvent ( nil, eventType, touch )
        self._prevTouch = touch
    end
end

return GestureRecognizer
