local core = require ( "core" )

local BoardModel = core.class ( Object )
{
    R_BULLS_EYE = 1,
    R_BULL = 2,
    R_INNER_RING = 3,
    R_TRIPLE_RING = 4,
    R_OUTER_RING = 5,
    R_DOUBLE_RING = 6,
    R_FULL = 7,
    R_NONE = 8,
}

local M_2PI = 2 * math.pi
local DEFAULT_BULLS_EYE_VAL = 50
local DEFAULT_BULL_VAL = 25
local DEFAULT_TRIPLE_VAL = 3
local DEFAULT_DOUBLE_VAL = 2

function BoardModel:setRadiuses ( radiuses, bullsEyeVal, bullVal, tripleRingSF, doubleRingSF )
    local res = {}
    
    local setCircle = function ( idx, val )
        res [ idx ] = { value = val, r = 0 }
    end
    
    local bullsEyeVal = bullsEyeVal or DEFAULT_BULLS_EYE_VAL
    local bullVal = bullVal or DEFAULT_BULL_VAL
    local tripleRingSF = tripleRingSF or DEFAULT_TRIPLE_VAL
    local doubleRingSF = doubleRingSF or DEFAULT_DOUBLE_VAL
    
    setCircle ( BoardModel.R_BULLS_EYE, bullsEyeVal )
    setCircle ( BoardModel.R_BULL, bullVal )
    setCircle ( BoardModel.R_INNER_RING, 1 )
    setCircle ( BoardModel.R_TRIPLE_RING, tripleRingSF )
    setCircle ( BoardModel.R_OUTER_RING, 1 )
    setCircle ( BoardModel.R_DOUBLE_RING, doubleRingSF )
    setCircle ( BoardModel.R_FULL, 0 )
    setCircle ( BoardModel.R_NONE, 0 )
    
    for i = 1, #radiuses do
        res [ i ].r  = radiuses [ i ]
    end
    
    self._circles =  res
end

function BoardModel.isDouble ( circle )
    return circle == BoardModel.R_DOUBLE_RING or circle == BoardModel.R_BULLS_EYE
end

function BoardModel:getRFForValue ( value, circle )
    if circle <= BoardModel.R_BULL or circle >= BoardModel.R_FULL then
        return nil
    end
    local sectors = self._sectors
    local scaleCircleValue = self._circles [ circle ].value
    local n = #sectors
    local sIndex = nil
    for i = 2, n do
        local sector = sectors [ i ]
        if sector.value * scaleCircleValue == value then
            sIndex = i
            break
        end
    end
    if sIndex then
        return { 
            rMin = self._circles [ circle - 1 ].r, 
            rMax = self._circles [ circle ].r,
            fMin = sectors [ sIndex - 1 ].f,
            fMax = sectors [ sIndex ].f,
            }
    end
end

function BoardModel:getDoubleOrSingleRF ( value )
    return self:getRFForValue ( value, BoardModel.R_DOUBLE_RING ) or self:getRFForValue ( value, BoardModel.R_OUTER_RING )
end

function BoardModel:getDoubleRF ( value )
    return self:getRFForValue ( value, BoardModel.R_DOUBLE_RING )
end

function BoardModel:setSectors ( sectors )
    local n = #sectors
    local delta = 2 * math.pi / n

    local currF = delta * 0.5
    local res = {}
    for i = 0, n do
        local sector = {}
        sector.value = sectors [ i % n + 1 ]
        sector.f = currF
        res [ i + 1 ] = sector
        currF = math.min ( currF + delta, M_2PI )
    end
    
    self._sectors = res
end

function BoardModel.boardCoordinates ( x, y )
    return math.sqrt ( x * x + y * y ), ( math.atan2 ( y, x ) + M_2PI ) % M_2PI
end

function BoardModel.modelCoordinates ( r, f )
    return r * math.cos ( f ), r * math.sin ( f )
end

function BoardModel:getCircle ( circle )
    return self._circles [ circle ]
end

function BoardModel:circles ()
    return self._circles
end

function BoardModel:sectors ()
    return self._sectors
end

function BoardModel:getSectorByValue ( value )
    local sectors = self._sectors
    local n = #sectors
    local sn = nil
    local sector = nil
    for i = 1, n do
        if sectors [ i ].value == value then
            sn = i
            sector = sectors [ i ]
            break
        end
    end
    return sn, sector
end

function BoardModel:getTargetedValue ( x, y )
    self.lastTarget = { x, y }
    local r, f = BoardModel.boardCoordinates ( x, y )
    local circles = self._circles
    local sectors = self._sectors
    
    local sn = 0
    for i = 1, #sectors do
        if f < sectors [ i ].f then
            sn = i
            break
        end
    end
    
    local cn = BoardModel.R_NONE
    for i = 1, BoardModel.R_FULL do
        if r < circles [ i ].r then
            cn = i 
            break
        end
    end
    
    local value = 0
    local cValue = circles [ cn ].value
    local sValue = 0
    if cn <= BoardModel.R_BULL then
        value = cValue
        sValue = cValue
    else
        if cn < BoardModel.R_FULL then
            sValue = sectors [ sn ].value
        end
        value = sValue * cValue
    end
    
    return value, cn, sValue
end

return BoardModel
