local core = require ( "core" )

local PlayerModel = require ( "Model.PlayerModel" )
local BoardModel = require ( "Model.BoardModel" )

local Events = require ( "Events" )
local ModelEvents = Events.E_MODEL

local GameModel = core.class ( Object )
{
    STATE_NONE = "STATE_NONE",
    STATE_INIT = "STATE_INIT",
    STATE_ROUND_BEG = "STATE_ROUND_BEG",
    STATE_TURN_BEG = "STATE_TURN_BEG",
    STATE_TURN = "STATE_TURN",
    STATE_TURN_COMPLETE = "STATE_TURN_COMPLETE",
    STATE_TURN_END = "STATE_TURN_END",
    STATE_ROUND_COMPLETE = "STATE_ROUND_COMPLETE",
    STATE_ROUND_END = "STATE_ROUND_END",
    STATE_DONE = "STATE_DONE",
    
    PLAYER_1 = 1,
    PLAYER_2 = 2,
    
    HIT_COMMON = 1,
    HIT_BUST = 2,
    HIT_WIN = 3,
    HIT_TURN = 4,
}

function GameModel:ctor ()
    self._players = {}
    self._state = GameModel.STATE_NONE
end

function GameModel:_switchState ( nextState )
    local prevState = self._state
    self._state = nextState
    -- log ( "Model.state %s -> %s", prevState, nextState )
    app.queue:postEvent ( self, ModelEvents.STATE_CHANGED, { prevState = prevState, nextState = nextState })
end

function GameModel:setup ( rounds, singlePlayer )
    assert ( self._state == GameModel.STATE_NONE )
    self._players [ GameModel.PLAYER_1 ] = PlayerModel.new ( GameModel.PLAYER_1 )
    if not singlePlayer then
        self._players [ GameModel.PLAYER_2 ] = PlayerModel.new ( GameModel.PLAYER_2 )
    end
    
    local bm = BoardModel.new ()
    bm:setRadiuses ({ 6.0, 11.5, 70, 77.0, 116.0, 123, 150.0 })
    bm:setSectors ({ 6, 13, 4, 18, 1, 20, 5, 12, 9, 14, 11, 8, 16, 7, 19, 3, 17, 2, 15, 10 })
    self._board = bm
    
    self._roundsRemain = rounds or 1
    
    self:_switchState ( GameModel.STATE_INIT )
end

function GameModel:start ( score, firstPlayer )
    assert ( self._state == GameModel.STATE_INIT or self._state == GameModel.STATE_ROUND_END )
    self:_roundSetup ( score, firstPlayer )
    self:_switchState ( GameModel.STATE_ROUND_BEG )
end

function GameModel:_roundSetup ( score, firstPlayer )
    for i, pl in ipairs ( self._players ) do
        pl:startSetup ( score )
    end
    local activePlayer = self._activePlayer
    if activePlayer and activePlayer:isActive () then
        activePlayer:deactivate ()
    end
    local firstPlayer = firstPlayer or GameModel.PLAYER_1
    activePlayer = self._players [ firstPlayer ]
    activePlayer:activate ()
    self._activePlayer = activePlayer
end

function GameModel:startRound ()
    assert ( self._state == GameModel.STATE_ROUND_BEG )
    self:_switchState ( GameModel.STATE_TURN_BEG )
    self._roundsRemain = self._roundsRemain - 1
end

function GameModel:startTurn ()
    assert ( self._state == GameModel.STATE_TURN_BEG )
    self:_switchState ( GameModel.STATE_TURN )
end

function GameModel:forceCompleteTurn ()
    assert ( self._state == GameModel.STATE_TURN )
    self:_switchState ( GameModel.STATE_TURN_COMPLETE )
end

function GameModel:finishTurn ()
    assert ( self._state == GameModel.STATE_TURN_COMPLETE )
    self:_switchState ( GameModel.STATE_TURN_END )
end

function GameModel:nextTurn ()
    assert ( self._state == GameModel.STATE_TURN_END )
    
    local activePlayer = self._activePlayer
    activePlayer:deactivate ()
    local players = self._players
    activePlayer = players [ activePlayer:index () % #players + 1 ]
    activePlayer:activate ()
    self._activePlayer = activePlayer
    
    self:_switchState ( GameModel.STATE_TURN_BEG )
end

function GameModel:finishRound ()
    assert ( self._state == GameModel.STATE_ROUND_COMPLETE )
    local activePlayer = self._activePlayer
    if activePlayer and activePlayer:isActive () then
        activePlayer:deactivate ()
    end
    self._activePlayer = nil
    for i, pl in ipairs ( self._players ) do
        pl:cleanup ()
    end
    if self._roundsRemain > 0 then
        self:_switchState ( GameModel.STATE_ROUND_END )
    else
        self:_switchState ( GameModel.STATE_DONE )
    end
end

function GameModel:nextRound ( score )
    self:start ( score )
end

function GameModel:processHit ( x, y )
    assert ( self._state == self.STATE_TURN )
    local activePlayer = self._activePlayer
    local value, circle, sValue = self._board:getTargetedValue ( x, y )
    local aV, aT = activePlayer:processHit ( value )
    
    local result = GameModel.HIT_COMMON
    if aV < 0 --[[or aV == 1--]] then
        -- bust
        self:_switchState ( GameModel.STATE_TURN_COMPLETE )
        activePlayer:recalcScore ( false )
        result = GameModel.HIT_BUST
    elseif aV == 0 then
        if true or BoardModel.isDouble ( circle ) then
            -- game over
            self:_switchState ( GameModel.STATE_ROUND_COMPLETE )
            activePlayer:recalcScore ( true )
            result = GameModel.HIT_WIN
        else
            -- bust
            self:_switchState ( GameModel.STATE_TURN_COMPLETE )
            activePlayer:recalcScore ( false )
            result = GameModel.HIT_BUST
        end
    elseif aT <= 0 then
        -- no darts
        self:_switchState ( GameModel.STATE_TURN_COMPLETE )
            activePlayer:recalcScore ( true )
        result = GameModel.HIT_TURN
    end
    return result, value, circle, sValue
end

function GameModel:activePlayer ()
    return self._activePlayer
end

function GameModel:players ()
    return self._players
end

function GameModel:state ()
    return self._state
end

function GameModel:board ()
    return self._board
end

return GameModel
