local _NAME = ...
local GestureRecognizer = require ( "Input.GestureRecognizer" )
local Events = require ( "Events" )
local PlayerView = require ( "View.PlayerView" )
local ScoreView = require ( "View.ScoreView" )
local BoardView = require ( "View.BoardView" )
local AnimEvents = Events.E_ANIMATION
local GameEvents = Events.E_GAME
local ModelEvents = Events.E_MODEL
local Stack = require ( "Stack" )
local MessageBox = require ( "View.MessageBox" )
local InfoScreen = require ( "View.InfoScreen" )
local SoundManager = require ( "SoundManager" )

local GameView = core.class ( Object )
{
    VIEW_SIZE = { width = 320, height = 480 },
    MIN_DARTS_VELOCITY = 5,
    MAX_DARTS_ON_BOARD = 3,
    TOUCH_SMOOTHING = 3,
    
    GESTURES = 
    {    
        NONE = 1,
        TAP = 2,
        PAN = 3,
        TAP_PAN = 4,
    };
    
    DARTS_PIVOT = { x = 165, y = 75 },
    THROW_VELOCITY_OFFSET = { x = 0, y = -25 },
    VELOCITY_SCALE_FACTOR = 0.1,
    
    PRIOR_BOTTOM = 0,
    PRIOR_MESSAGE_BOX = 5,
    PRIOR_CONFIRM_BOX = 10,
    PRIOR_TOP = 100,
}
GameView.DARTS_MODEL_OFFSET = { x = 155 - GameView.DARTS_PIVOT.x, y = 169 - GameView.DARTS_PIVOT.y }

function GameView:ctor ( resource, actions )
    local screenWidth, screenHeight = MOAIGfxDevice.getViewSize ()
    screenWidth, screenHeight = screenWidth or 0, screenHeight or 0
    local screenType = 0
    if screenWidth == 0 or screenHeight == 0 then
        if screenType == 1 then
            screenWidth, screenHeight = 640 * 0.5, 1136 * 0.5
        elseif screenType == 2 then
            screenWidth, screenHeight = 768, 1024
        elseif screenType == 3 then
            screenWidth, screenHeight = GameView.VIEW_SIZE.width * 2.0, GameView.VIEW_SIZE.height * 2.0
        else
            screenWidth, screenHeight = GameView.VIEW_SIZE.width, GameView.VIEW_SIZE.height
        end
    end
    
    MOAISim.openWindow ( "GameView", screenWidth, screenHeight )

    self:resize ( screenWidth, screenHeight, true )
    
    self.renderTable = {}

    local resource = resource or require ( _NAME .. ".resources" )
    for _, res in pairs ( resource ) do
        res:create ( self )
    end
    
    local actions = actions or require ( _NAME .. ".actions" )
    self.actions = actions
    
    local gr = GestureRecognizer.new ( self )
    gr:initInput ()
    self.gestureRecognizer = gr
    
    self.sigDartsHit = core.signal ()
    self.sigDartsThrow = core.signal ()
    self.sigGesBegan = core.signal ()
    self.sigTapEnded = core.signal ()
    self.sigPanMoved = core.signal ()
    self.sigPanEnded = core.signal ()
    
    self._players = { PlayerView.new (), PlayerView.new ()}
    for i, pl in ipairs ( self._players ) do
        pl:setTextBox ( self [ "textBox_" .. i ])
        pl:setNameBox ( self [ "nameBox_" .. i ])
        pl:setTurnIndicator ( self [ "turnIndicator_" .. i ])
        local idx = 1
        local cnt = self [ "dartsCounter_" .. i .. idx ]
        while cnt do
            pl:addDartsCounter ( cnt )
            idx = idx + 1
            cnt = self [ "dartsCounter_" .. i .. idx ]
        end
    end
    
    local i = 1
    local textBox = self [ "scoreBox_" .. i ]
    self._scoreView = ScoreView.new ()
    while textBox do
        self._scoreView:addScoreBox ( textBox )
        i = i + 1
        textBox = self [ "scoreBox_" .. i ]        
    end
    
    self._inputPartitions = Stack.new ()
    self._confirmPopup = MessageBox.new ( self.popupLayer, GameView.PRIOR_CONFIRM_BOX )
    self._confirmPopup.popupsRoot:setParent ( self.mainRoot )

    self:setAllowedGestures ( GameView.GESTURES.NONE )
end

function GameView:resize ( screenWidth, screenHeight, firstTime )
    local viewport = self.viewport
    if not viewport then
        viewport = MOAIViewport.new ()
        self.viewport = viewport
    end
    viewport:setOffset ( -1, -1 )
    
    if firstTime then
        local scale = math.min ( screenWidth / GameView.VIEW_SIZE.width, screenHeight / GameView.VIEW_SIZE.height )
        local dx, dy = screenWidth / scale - GameView.VIEW_SIZE.width, screenHeight / scale -  GameView.VIEW_SIZE.height
        
        RealViewSize = { width = GameView.VIEW_SIZE.width + dx, height = GameView.VIEW_SIZE.height + dy }
    end
    if screenWidth < screenHeight then
        viewport:setScale ( RealViewSize.width, RealViewSize.height )
        if self.inputRoot then
            self.inputRoot:setRot ( 0 )
            self.mainRoot:setLoc ( 0, 0 )
        end
    else
        viewport:setScale ( RealViewSize.height, RealViewSize.width )
        if self.inputRoot then
            self.inputRoot:setRot ( 90 )
            self.mainRoot:setLoc ( 0, -RealViewSize.height )
        end
    end
    viewport:setSize ( 0, 0, screenWidth, screenHeight )
end

function GameView:setModel ( model )
    self._model = model
    local mPl = model:players ()
    local vPl = self._players
    for i = 1, math.min ( #mPl, #vPl ) do
        vPl [ i ]:setPlayerModel ( mPl [ i ])
    end
end

function GameView:showSplashScreen ()
    self:_pushPartitions ()
    MOAIRenderMgr.setRenderTable
    {
        self.gameBoardLayer,
        self.gameDartsLayer,
        self.splashScreenLayer,
        self.playerHUDLayer,
        self.popupLayer,
    }
    self:setAllowedGestures ( GameView.GESTURES.NONE )
    self.actions.SplashAnim:create ( self ):start ()
end

function GameView:showGameScreen ()
    self:_pushPartitions ( self.playerHUDLayer:getPartition ())
    MOAIRenderMgr.setRenderTable { self.gameBoardLayer, self.gameDartsLayer, self.playerHUDLayer, self.popupLayer }
    self:setAllowedGestures ( GameView.GESTURES.TAP )
end

function GameView:displayTurnLabel ( index, text )
    if index == 1 then
        self.turnLabel:setAlignment ( MOAITextBox.LEFT_JUSTIFY, MOAITextBox.RIGHT_JUSTIFY )
    else
        self.turnLabel:setAlignment ( MOAITextBox.RIGHT_JUSTIFY, MOAITextBox.RIGHT_JUSTIFY )
    end
    self.turnLabel:setString ( text )
end

function GameView:hideTurnLabel ()
    self.turnLabel:setString ( "" )
end

function GameView:_recalcAllowPan ()
    local allowPan = false
    for _, part in ipairs ( self._inputPartitions:top ()) do
        if part.panAllowed then
            allowPan = true
            break
        end
    end
    self._topPartitionAllowPanGesture = allowPan
end

function GameView:_pushPartitions ( ... )
    self._inputPartitions:push { ... }
    self:_recalcAllowPan ()
end

function GameView:_popPartitions ()
    self._inputPartitions:pop ()
    self:_recalcAllowPan ()
end

function GameView:boardCenter ()
    return self.gameBoard:getLoc ()
end
-------------- turn controll -------------

function GameView:dartsAiming ( x, y )
    local x = x or RealViewSize.width * 0.5
    local y = y or 216
    local darts = self._currentDarts
    local create = false
    if not darts then
        darts = self:createDarts ( self._model:activePlayer ():index ())
        self._currentDarts = darts
        app.queue:postEvent ( self._model:activePlayer (), AnimEvents.DARTS_READY )
        create = true
    end
    darts:setLoc ( x, y )
    return create
end

function GameView:endDartsAiming ( vx, vy )
    local darts = self._currentDarts
    if darts then
        if vy >= GameView.MIN_DARTS_VELOCITY then
            local x, y = darts:getLoc ()
            local mv = math.sqrt ( vx * vx + vy * vy )
            self:throwDarts ( darts, vx, vy, mv )
            app.queue:subscribe ( self, GameView.onDartsHit, AnimEvents.DARTS_ANIM )
            self.sigDartsThrow ( x, y, vx, vy )
        else
            self.playerHUDLayer:removeProp ( darts )
            app.queue:postEvent ( self._model:activePlayer (), AnimEvents.DARTS_DROPPED )
            --self:dartsAiming ()
        end
        self._currentDarts = nil        
    end 
end

function GameView:boardModelToWorld ( x, y )
    local bx, by = self.gameBoard:getLoc ()
    local x, y = ( x - GameView.DARTS_MODEL_OFFSET.x ) + bx, ( y - GameView.DARTS_MODEL_OFFSET.y ) + by
    return x, y
end

function GameView:modelVelocityToWorld ( vx, vy )
    return vx - GameView.THROW_VELOCITY_OFFSET.x / GameView.VELOCITY_SCALE_FACTOR, vy - GameView.THROW_VELOCITY_OFFSET.y / GameView.VELOCITY_SCALE_FACTOR
end

function GameView:finalModelCoordinates ( x, y, vx, vy )
    local bx, by = self.gameBoard:getLoc ()
    local x, y = ( x + GameView.DARTS_MODEL_OFFSET.x ) - bx, ( y + GameView.DARTS_MODEL_OFFSET.y ) - by
    local vx, vy = vx * GameView.VELOCITY_SCALE_FACTOR + GameView.THROW_VELOCITY_OFFSET.x, vy * GameView.VELOCITY_SCALE_FACTOR + GameView.THROW_VELOCITY_OFFSET.y
    return x + vx, y + vy
end

function GameView:onDartsHit ( darts )
    app.queue:unsubscribe ( self, GameView.onDartsHit, AnimEvents.DARTS_ANIM )
    
    local x, y = darts:getLoc ()
    local bx, by = self.gameBoard:getLoc ()
    x, y = ( x + GameView.DARTS_MODEL_OFFSET.x ) - bx, ( y + GameView.DARTS_MODEL_OFFSET.y ) - by
    local r = math.sqrt ( x * x + y * y )
    local w, h = self.gameBoardQ:getDims ()
    
    local allDarts = self._allDarts
    if not allDarts then
        allDarts = { start = 1, maxSize = GameView.MAX_DARTS_ON_BOARD }
        self._allDarts = allDarts
    end
    local st = allDarts.start
    local old = allDarts [ st ]
    allDarts [ st ] = darts
    st = st % allDarts.maxSize + 1
    allDarts.start = st
    if old then
        self.gameDartsLayer:removeProp ( old )
    end
    
    self.gameDartsLayer:insertProp ( darts )
    self.sigDartsHit ( x, y )
end

function GameView:higlightSector ( sector, circle )
    self.sectorHighlight:setRot ( sector and ( sector - 1 ) * 18 - 90 or 0)
    self.sectorHighlight:setIndex ( circle )
    self.sectorHighlight:setVisible ( true )
end

function GameView:unhiglightSector ()
    self.sectorHighlight:setVisible ( false )
end

function GameView:highlightDarts ()
    self:unhighlightDarts ()
    local darts = self._currentDarts
    if darts then
        local dartsHiglight = self:createDartsHighlight ()
        dartsHiglight:setParent ( darts )
        self._dartsHighlight = dartsHiglight
    end
end

function GameView:unhighlightDarts ()
    local dartsHiglight = self._dartsHighlight
    if dartsHiglight then
        self.playerHUDLayer:removeProp ( dartsHiglight )
        self._dartsHighlight = nil
    end
end

function GameView:setFaderVisible ( visible, a, length )
    local fadersCount = self._fadersCount or 0
    local a = a or 0.5
    local length = length or 0.5
    if visible then
        if fadersCount == 0 then
            if self._faderAnim then
                self._faderAnim:stop ()
                self._faderAnim = nil
            end
            self._faderAnim = self.actions.FadeColor:create ( self.fader, { r = 0, g = 0, b = 0, a = a, length = length }):start ()
        end
        fadersCount = fadersCount + 1
    else
        if fadersCount == 1 then
            if self._faderAnim then
                self._faderAnim:stop ()
                self._faderAnim = nil
            end
            self._faderAnim = self.actions.FadeColor:create ( self.fader, { r = 0, g = 0, b = 0, a = 0.0, length = length }):start ()
        end
        fadersCount = fadersCount - 1
    end
    self._fadersCount = fadersCount
end

function GameView:showInfo ()
    self:setFaderVisible ( true, nil, 0.1 )
    
    local infoScreen = InfoScreen.new ( self.viewport )
    infoScreen.root:setParent ( self.mainRoot )
    infoScreen.root:scheduleUpdate ()
    self:_pushPartitions ( infoScreen.mainLayer:getPartition ())
    local renderTable = MOAIRenderMgr.getRenderTable ()
    renderTable [ #renderTable + 1 ] = infoScreen.mainLayer
    MOAIRenderMgr.setRenderTable ( renderTable )
    self._infoScreen = infoScreen
    
    app.queue:subscribe ( self, GameView.hideInfo, GameEvents.HIDE_INFO )
    self.fader:setPriority ( GameView.PRIOR_TOP )
    self.hintText:setVisible ( false )
end

function GameView:hideInfo ()
    self.hintText:setVisible ( true, nil, 0.05 )
    self:setFaderVisible ( false )
    self.fader:setPriority ( GameView.PRIOR_BOTTOM )
    app.queue:unsubscribe ( self, GameView.hideInfo, GameEvents.HIDE_INFO )
    
    local infoScreen = self._infoScreen
    local renderTable = MOAIRenderMgr.getRenderTable ()
    renderTable [ #renderTable ] = nil
    MOAIRenderMgr.setRenderTable ( renderTable )
    self:_popPartitions ()
    self._infoScreen = nil
end


function GameView:showHint ( text )
    self.hintText:seekColor ( 1, 1, 1, 1, 0.3 )
    self.hintText:setString ( text )
end

function GameView:hideHint ()
    self.hintText:seekColor ( 1, 1, 1, 0, 0.3 )
    local timer = MOAITimer.new ()
    timer:setSpan ( 0.3 )
    timer:setMode ( MOAITimer.NORMAL )
    timer:setListener ( MOAITimer.EVENT_TIMER_END_SPAN,
        function ()
            self.hintText:setString ( "" )
        end
    )
    timer:start ()
end

function GameView:setHintAlpha ( alpha )
    self.hintText:seekColor ( 1, 1, 1, alpha, 0.25 )
end

function GameView:showSwipeHint ()
    self.hintSwipe:setVisible ( true )
    self._hintSwipeAction = self.actions.HintSwipe:create ( self )
    self._hintSwipeAction:start ()
end

function GameView:hideSwipeHint ()
    self.hintSwipe:setVisible ( false )
    self._hintSwipeAction = nil
end

function GameView:displayConfirmPopup ( text, buttonLeft, buttonRight )
    self:_pushPartitions ( self.popupLayer:getPartition ())
    local confirmPopup = self._confirmPopup
    confirmPopup:setText ( text )
    confirmPopup:setButtonText ( MessageBox.BUTTON_LEFT, buttonLeft.text )
    confirmPopup:setButtonText ( MessageBox.BUTTON_RIGHT, buttonRight.text )
    confirmPopup:setButtonEvent ( MessageBox.BUTTON_LEFT, buttonLeft.event )
    confirmPopup:setButtonEvent ( MessageBox.BUTTON_RIGHT, buttonRight.event )
    confirmPopup:show ( true, true )
    self:setFaderVisible ( true )
end

function GameView:displayTextPopup ( text, info, idx )
    self:_pushPartitions ( self.popupLayer:getPartition ())
    
    local ap = self._model and self._model:activePlayer ()
    local idx = idx or ( ap and ap:index ()) or 1
    self._messageBox = MessageBox.new ( self.popupLayer, GameView.PRIOR_MESSAGE_BOX )
    self._messageBox.popupsRoot:setParent ( self.mainRoot )
    self._messageBox:setIndex ( idx )
    self._messageBox:setText ( text )
    if info then
        self._messageBox:setInfoText ( info )
    end
    self._messageBox:show ( false, true )
end

function GameView:changeTextPopup ( text )
    if self._messageBox:isVisible () then
        self._messageBox:setText ( text )
    end
end
  
function GameView:hideTextPopup ()
    if self._messageBox and self._messageBox:isVisible () then
        self._messageBox:hide ( true )
        self._messageBox = nil
        self:_popPartitions ()
    end
end

function GameView:hideConfirmPopup ()
    if self._confirmPopup:isVisible () then
        self._confirmPopup:hide ( true )
        self:_popPartitions ()
        self:setFaderVisible ( false )
    end
end
  
-------------- darts processing ----------

function GameView:createDarts ( index )
    local darts = MOAIProp2D.new ()
    darts:setParent ( self.mainRoot )
    darts:setDeck ( self.dartsDecks [ index ])
    darts:setPiv ( GameView.DARTS_PIVOT.x, GameView.DARTS_PIVOT.y )
    self.playerHUDLayer:insertProp ( darts )
    return darts
end

function GameView:createDartsHighlight ()
    local dartsHighlight = MOAIProp2D.new ()
    dartsHighlight:setParent ( self.mainRoot )
    dartsHighlight:setDeck ( self.dartsHighlightDeck )
    self.playerHUDLayer:insertProp ( dartsHighlight )
    dartsHighlight:setPriority ( 20 )
    return dartsHighlight
end

-------------- animations ----------------

function GameView:endTurnAnimate ()
    self.actions.EndTurnAnim:create ():start ()
end

function GameView:clearBoard ()
    local allDarts = self._allDarts
    if allDarts then
        for _, d in ipairs ( allDarts ) do
            self.gameDartsLayer:removeProp ( d )
        end
    end
    self._allDarts = nil
    self._scoreView:clearScores ()
end

function GameView:displayScore ( value, sValue, circle )
    self._scoreView:displayNextScore ( value, sValue, circle )
end

function GameView:throwDarts ( darts, vx, vy, length )
    local length = 0.5 --math.min ( 0.5, math.max ( 0.4, 20.0 / length ))
    local dx = vx * GameView.VELOCITY_SCALE_FACTOR + GameView.THROW_VELOCITY_OFFSET.x
    local dy = vy * GameView.VELOCITY_SCALE_FACTOR + GameView.THROW_VELOCITY_OFFSET.y
    self.actions.ThrowDarts:create ( darts, { dx = dx, dy = dy, length = length }):start ()
end

-------------- input processing ----------------

function GameView:enableInput ()
    self.gestureRecognizer:enableInput ()
end

function GameView:disableInput ()
    self.gestureRecognizer:disableInput ()
end

function GameView:setAllowedGestures ( gesture )
    local gBegan, gTapE, gPanM, gPanE = nil, nil, nil, nil
    if gesture == GameView.GESTURES.TAP then
        gBegan = GameView.handleGestureBegan
        gTapE = GameView.handleTap
    elseif gesture == GameView.GESTURES.PAN then
        gBegan = GameView.handleGestureBegan
        gPanM = GameView.handlePanMove
        gPanE = GameView.handlePanEnd
    elseif gesture == GameView.GESTURES.TAP_PAN then
        gBegan = GameView.handleGestureBegan
        gTapE = GameView.handleTap
        gPanM = GameView.handlePanMove
        gPanE = GameView.handlePanEnd
    end
    local gr = self.gestureRecognizer
    gr:setCallback ( GestureRecognizer.GR_GESTURE_BEGAN, gBegan )
    gr:setCallback ( GestureRecognizer.GR_TAP_ENDED, gTapE )
    gr:setCallback ( GestureRecognizer.GR_PAN_MOVED, gPanM )
    gr:setCallback ( GestureRecognizer.GR_PAN_ENDED, gPanE )
end

local function convertTouch ( layer, root, touch )
    local rx, ry, timestamp = unpack ( touch )
    rx, ry = layer:wndToWorld ( rx, ry )
    local x, y = root:worldToModel ( rx, ry )
    return x, y, rx, ry
end

local function createTouchContainer ( maxSize )
    return { start = 1, maxSize = maxSize }
end

local function getMeanVelocity ( tc )
    local vx, vy = 0, 0
    local n = math.min ( tc.maxSize, #tc )
    local st = tc.start
    st = ( st - 1 ) % n + 1
    local prevTouch = tc [ st ]
    local currTouch = nil
    for i = 0, n - 2 do
        currTouch = tc [( st + i ) % n + 1 ]
        vx = vx + ( currTouch.x - prevTouch.x )
        vy = vy + ( currTouch.y - prevTouch.y )
        prevTouch = currTouch
    end
    if n > 1 then
        local k = 1 / ( n - 1 )
        vx, vy = vx * k, vy * k
    end
    return vx, vy
end

local function addTouchToContainer ( tc, x, y )
    local st = tc.start
    tc [ st ] = { x = x, y = y }
    st = st % tc.maxSize + 1
    tc.start = st
end

local function trackTouch ( target, x, y, isInside )
    if not target then
        return
    end

    if isInside == nil then
        isInside = target:inside ( x, y )
    end

    local handler
    if isInside then
        handler = target.onTouchEnter 
    else
        handler = target.onTouchLeave
    end

    if handler then
        handler ( target )
    end
end

function GameView:handleGestureBegan ( _, touch )
    self._prevTouches = createTouchContainer ( GameView.TOUCH_SMOOTHING )
    local x, y, rx, ry = convertTouch ( self.playerHUDLayer, self.inputRoot, touch )
    addTouchToContainer ( self._prevTouches, x, y )
    local alowCreate = self._topPartitionAllowPanGesture
    for _, part in ipairs ( self._inputPartitions:top ()) do
        local props = { part:propListForPoint ( rx, ry )}
        for k, prop in pairs ( props ) do
            if prop.isClickable then
                self._startProp = prop
                alowCreate = nil
                trackTouch ( self._startProp, x, y )
                break
            end
        end
    end
    self.sigGesBegan ( x, y, alowCreate )
end

function GameView:handleTap ( _, touch )
    local x, y, rx, ry = convertTouch ( self.playerHUDLayer, self.inputRoot, touch )
    
    local prop = self._startProp
    local processed = false

    if prop then
        trackTouch ( prop, x, y, false )
        if prop:inside ( x, y ) then
            if prop.onClick then
                prop:onClick ()
                processed = true
            end
        end
    end

    if not processed then
        app.queue:postEvent ( nil, GameEvents.DEFAULT_TAP )
    end
    
    self._startProp = nil
    self.sigTapEnded ( x, y )
end

function GameView:handlePanMove ( state, touch )
    local x, y, rx, ry = convertTouch ( self.playerHUDLayer, self.inputRoot, touch )

    if self._topPartitionAllowPanGesture then
        addTouchToContainer ( self._prevTouches, x, y )
        self.sigPanMoved ( x, y )
    end

    trackTouch ( self._startProp, x, y )
end

function GameView:handlePanEnd ( state, touch )
    local x, y, rx, ry = convertTouch ( self.playerHUDLayer, self.inputRoot, touch )
    if self._topPartitionAllowPanGesture then
        local vx, vy = getMeanVelocity ( self._prevTouches )   
        self.sigPanEnded ( x, y, vx, vy )
    end

    local prop = self._startProp
    if prop then
        trackTouch ( prop, x, y, false )
        if prop:inside ( x, y ) then
            if prop.onClick then
                prop:onClick ()
            end
        end
    end

    self._startProp = nil
end

function GameView:canGrabDart ( x, y )
    return y <= RealViewSize.height - 50
end

return GameView
