local core = require ( "core" )

local PlayerView = core.class ( Object )
{
    INDICATOR_ACTIVE = 1,
    INDICATOR_INACTIVE = 2,
}

local Events = require ( "Events" )
local PlayerEvents = Events.E_PLAYER
local AnimEvents = Events.E_ANIMATION
local ModelEvents = Events.E_MODEL

local GameModel = require ( "Model.GameModel" )

function PlayerView:setPlayerModel ( model )
    if self._player then
        app.queue:unsubscribe ( self, PlayerView.onPlayerModelChanged, PlayerEvents, self._player )
    else
        app.queue:subscribe ( self, PlayerView.onPlayerModelChanged, PlayerEvents, model )
        app.queue:subscribe ( self, PlayerView.onDartsReady, AnimEvents.DARTS_READY, model )
        app.queue:subscribe ( self, PlayerView.onDartsDropped, AnimEvents.DARTS_DROPPED, model )
        app.queue:subscribe ( self, PlayerView.onGameModelChanged, ModelEvents.STATE_CHANGED )
    end
    self._player = model
end

function PlayerView:setNameBox ( namebox )
    self._namebox = namebox
end

function PlayerView:setTurnIndicator ( indicator )
    self._turnIndicator = indicator
end

function PlayerView:addDartsCounter ( counter )
    local counters = self._counters
    if not counters then
        counters = {}
        self._counters = counters
    end
    counters [ #counters + 1 ] = counter
end

function PlayerView:clearDartsCounters ()
    self._counters = nil
end

function PlayerView:setTextBox ( textbox )
    self._textbox = textbox
end

function PlayerView:onDartsReady ( player, event )
    local throws = player:availableThrows ()
    self:_displayThrows ( throws - 1 )
end

function PlayerView:onDartsDropped ( player, event )
    local throws = player:availableThrows ()
    self:_displayThrows ( throws )
end

function PlayerView:onPlayerModelChanged ( player, event )
    if event == PlayerEvents.SCORE_CHANGED then
        self:_displayScore ( player )
    elseif event == PlayerEvents.ACTIVATED then
        self:_displayIndicator ( player )
        self:_displayColorName ( player )
        self:_startActivateAnim ( player )
    elseif event == PlayerEvents.DEACTIVATED then
        self:_displayIndicator ( player )
        self:_displayColorName ( player )
        self:_stopActivateAnim ( player )
    elseif event == PlayerEvents.THROWS_CHANGED then
        self:_displayAvailableThrows ( player )
    end
end

function PlayerView:_displayAvailableThrows ( player )
    local throws = player:availableThrows ()
    self:_displayThrows ( throws )
end

function PlayerView:_displayThrows ( throws )
    local counters = self._counters
    if counters then
        for i, c in ipairs ( counters ) do
            c:setVisible ( i <= throws )
        end
    end
end

function PlayerView:_startActivateAnim ( player )
    local counters = self._counters
    if counters then
        for i, c in ipairs ( counters ) do
            c:moveLoc ( 0, 15, 0.5 )
            c:seekScl ( 1.2, 1.2, 0.5 )
        end
    end
end

function PlayerView:_stopActivateAnim ( player )
    local counters = self._counters
    if counters then
        for i, c in ipairs ( counters ) do
            c:moveLoc ( 0, -15, 0.5 )
            c:seekScl ( 1.0, 1.0, 0.5 )
        end
    end
end

function PlayerView:_displayIndicator ( player )
    local indicator = self._turnIndicator
    if indicator then
        indicator:setIndex ( player:isActive () and PlayerView.INDICATOR_ACTIVE or PlayerView.INDICATOR_INACTIVE )
    end
end

function PlayerView:_displayScore ( player )
    local textbox = self._textbox
    if textbox then
        local score = player:currentScore ()
        textbox:setString ( score and tostring ( score ) or "" )
    end
end

function PlayerView:_displayName ( name )
    local namebox = self._namebox
    if namebox then
        namebox:setString ( name )
    end
end

function PlayerView:_colorName ( r, g, b )
    local namebox = self._namebox
    if namebox then
        namebox:setColor ( r, g, b )
    end
end

function PlayerView:_displayColorName ( player )
    local namebox = self._namebox
    if namebox then
        local r, g, b = 1, 1, 1
        if player:isActive () then
            if player:index () == GameModel.PLAYER_1 then
                r, g, b = 0.3, 0.5, 0.75
            else
                r, g, b = 1, 0, 0
            end
        end
        namebox:setColor ( r, g, b )
    end
end

function PlayerView:_updateName ()
    self:_displayName ( self._player:name ())
end

function PlayerView:onGameModelChanged ( sender, event, args )
    local model = sender
    if args.nextState == model.STATE_ROUND_BEG then
        self:_updateName ()
    elseif args.nextState == model.STATE_ROUND_END then
        self:_displayName ( "" )
    end
end

return PlayerView
