local _NAME = ...
local core = {}
_G[_NAME] = core
package.loaded[_NAME] = core

unpack = unpack or table.unpack

function log ( fmt, ... )
    local msg = string.format ( fmt, ... )
    print ( msg )
end

-- filter an array
core.filter = function ( t, predicate, inplace )
    local n = #t
    if inplace then
        for i = n, 1, -1 do
            if not predicate ( t [ i ] ) then
                table.remove ( t, i )
            end
        end
        return t
    else
        local filtered = {}
        for i = 1, n do
            local v = t [ i ]
            if predicate ( v ) then
                table.insert ( filtered, v )
            end
        end
        return filtered
    end
end;

core.map = function ( t, func, inplace )
    local mapped = inplace and t or {}
    local keys = {}
    for k, _ in pairs ( t ) do keys [ k ] = true end
    for k, _ in pairs ( keys ) do
        local v = t [ k ]
        mapped [ k ] = func ( v )
    end
    return mapped
end;

-- shallow copy
core.copy = function ( table )
    local copy = {}
    for k, v in pairs ( table ) do
        copy [ k ] = v
    end
    return copy
end;

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- classes
------------------------------------------------------------------------------------
core.class = function ( base )
    local imembers = base and setmetatable ({ super = base }, { __index = base.__imembers__ }) or {}
    local imt = { __index = imembers }

    local C = { __imembers__ = imembers }

    local function override ( cls, k, v )
        rawset ( cls, k, v )
        imembers [ k ] = v
    end

    local function statics ( cls, members )
        for k, v in pairs ( members ) do
            override ( C, k, v )
        end

        return C
    end

    local cmt =
    {
        __index = base,
        __newindex = override,
        __call = statics,
    }

    function C.alloc ( o )
        return setmetatable ( o or {}, imt )
    end

    function C.new ( ... )
        local o = C.alloc ()
        o = o:ctor ( ... ) or o
        return o
    end

    return setmetatable ( C, cmt )
end

Object = core.class ()

function Object:ctor ()
end

function Object:toString ()
    return tostring ( self )
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- observer pattern (synchronous callbacks)
------------------------------------------------------------------------------------
local function connect ( self, observer, method )
    local t = { o = observer, m = method }
    table.insert ( self, t )
end

local function disconnect ( self, observer, method )
    local n = #self
    for i = n, 1, -1 do
        if ( not observer or self [ i ].o == observer ) and ( not method or self [ i ].m == method ) then
            table.remove ( self, i )
        end
    end
end

local mt =
{
    __call = function ( self, ... )
        for _, callback in ipairs ( self ) do
            callback.m ( callback.o, ... )
        end
    end
}

local mt0 =
{
    __call = function ( self )
        for _, callback in ipairs ( self ) do
            callback.m ( callback.o )
        end
    end
}

local mt1 =
{
    __call = function ( self, arg1 )
        for _, callback in ipairs ( self ) do
            callback.m ( callback.o, arg1 )
        end
    end
}

function core.signal ()
    return setmetatable ({ connect = connect, disconnect = disconnect} , mt )
end

function core.signal0 ()
    return setmetatable ({ connect = connect, disconnect = disconnect }, mt0 )
end

function core.signal1 ()
    return setmetatable ({ connect = connect, disconnect = disconnect }, mt1 )
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- events
------------------------------------------------------------------------------------
Event = core.class ( Object )
{
    declare = function ( name )
        local e = Event.new ()

        e.class = e
        e._description = name or e

        e.types = function ( self, types )
            for k, v in pairs ( types ) do
                local subevent = Event.new ()
                subevent.class = e
                subevent._description = string.format ( "%s:%s", e._description, k )
                self [ k ] = subevent
            end
            return self
        end
        return e
    end;
}

function Event:toString ()
    return self._description
end

EventQueue = core.class ( Object )

function EventQueue:ctor ()
    self.processed = core.signal0 ()
    self._head = 0
    self._tail = 0
    self._byEvent   = {}
    self._bySender  = {}
    self.paused = false
end

function EventQueue:subscribe ( object, method, event, sender )
    assert ( object and method )
    assert ( event or sender )
    local handler = { o = object, m = method, s = sender, e = event }
    if event then
        local handlers = self._byEvent [ event.class ] or {}
        table.insert ( handlers, handler )
        self._byEvent [ event.class ] = handlers
    end
    if sender then
        local handlers = self._bySender [ sender ] or {}
        table.insert ( handlers, handler )
        self._bySender [ sender ] = handlers
    end
end

function EventQueue:unsubscribe ( object, method, event, sender )
    assert ( object )
    local handlers
    handlers = event and self._byEvent [ event.class ] or nil
    if handlers then
        core.filter ( handlers, function ( h )
            if h.o ~= object then return true end
            if not method then return false end
            return h.m ~= method
        end, true )
    end
    handlers = sender and self._bySender [ sender ] or nil
    if handlers then
        core.filter ( handlers, function ( h )
            if h.o ~= object then return true end
            if not sender then return false end
            return h.s ~= sender
        end, true )
    end
end

function EventQueue:postEvent ( sender, event, args )
    assert ( event )
    args = args or {}
    local head, tail = self._head, self._tail
    local msg = { s = sender, e = event, d = args }
    local cnt = tail - head

    self [ tail ] = msg
    self._tail = tail + 1

    if cnt == 0 and not self.paused then
        self:_process ()
    end
end

function EventQueue:_process ()
    local function invoke (h, s, e, d)
        if h.s and h.s ~= s then return nil end
        if h.e and h.e ~= e and h.e ~= e.class then return nil end
        h.m ( h.o, s, e, d )
        return true
    end

    local done = false
    local handlers, msg

    while not done do
        msg = self [ self._head ]

        local invoked = {}
        handlers = self._byEvent [ msg.e.class ]
        if handlers then
            for _, h in ipairs ( handlers ) do
                invoked [ h ] = invoked [ h ] or invoke ( h, msg.s, msg.e, msg.d )
            end
        end
        handlers = msg.s and self._bySender [ msg.s ] or nil
        if handlers then
            for _, h in ipairs ( handlers ) do
                invoked [ h ] = invoked [ h ] or invoke ( h, msg.s, msg.e, msg.d )
            end
        end

        -- remove message from the queue only after processing
        self._head = self._head + 1
        done = self._head == self._tail
    end
    self._head, self._tail = 0, 0
    self.processed ()
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- State machine
------------------------------------------------------------------------------------
FSM = core.class ( Object )

function FSM:ctor ( context, eventsIsNumbers )
    self._context = context
    self._locked = false
    
    if eventsIsNumbers then
        self._getHandler = function ( state, event )
            return state [ event ]
        end
    end
end

function FSM:declare ( states )
    self._states = states
    self._currState = states [ 1 ] or states.initial
    assert ( self._currState, "No initial state found" )
end

function FSM._getHandler ( state, event )
    return state [ event ] or state [ event.class ]
end

function FSM:_exec ( state, sender, event, args )
    assert ( event )
    assert ( not self._locked )
    self._locked = true

    local handler = self._getHandler ( state, event )

    local result
    if handler then
        result = handler ( self._context, sender, event, args )
        if type ( result ) == "string" or type ( result ) == "number" then
            local stateName = result
            result = self._states [ stateName ]
            assert ( result, "state not found: " .. stateName )
        end
    end
    self._locked = false
    return result
end

function FSM:handleEvent ( sender, event, args )
    local currState = self._currState
    assert ( event )
    assert ( currState )
    local nextState = self:_exec ( currState, sender, event, args )
    while nextState do
        self:_exec ( currState, self, "leave" )
        currState = nextState
        nextState = self:_exec ( currState, self, "enter" )
    end
    self._currState = currState
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- Resource manager
------------------------------------------------------------------------------------
rm =
{
    ID = "@ID";
}

-- declare resource
function rm:rsrc ( class, id )
    assert ( class )
    local rm = self
    return function ( cfg )
        local rsrc = cfg
        rsrc [ rm.ID ] = tostring ( id or rsrc )
        rsrc.create = function ( self, context, ... )
            return class:newFromRsrc( self, context, ... )
        end
        return rsrc
    end
end

function rm:find ( id, class )
    if class then
        local t = self._lookup [ class ]
        return t [ id ]
    end
    for class, t in pairs ( self._lookup ) do
        local rsrc = t [ id ]
        if rsrc then
            return rsrc, class
        end
    end
end
------------------------------------------------------------------------------------

ResourceCache = core.class ( Object )

function ResourceCache:ctor ( creator )
    self._lookup = setmetatable( {}, { __mode = 'v' })
    self._creator = creator
end

function ResourceCache:get ( key )
    local resource = self._lookup [ key ]
    if not resource then
        resource = self._creator ( key )
        self._lookup [ key ] = resource
    end
    return resource
end

function ResourceCache:set ( key, resource )
    self._lookup [ key ] = resource
end

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- application object
------------------------------------------------------------------------------------
E_SYSTEM = Event.declare "system" :types
{
    START = 1,
    QUIT = 2,
    SESSION_STARTED = 3,
}

E_TIMEOUT = Event.declare "timeout"
E_TICK = Event.declare "tick"

app =
{
    queue           = EventQueue.new ();
    tick            = core.signal1 ();

    _actions            = {},
    _completedActions   = {},
}

function app:onTick ( dt )
    self.queue.paused = true
    self.queue:postEvent ( nil, E_TICK, { dt = dt })
    self.tick ( dt )

    -- process all actions
    local completed = self._completedActions
    for action, _ in pairs ( self._actions ) do
        local done = action:update ( dt )
        if done then
            table.insert ( completed, action )
        end
    end

    while #completed > 0 do
        local action = table.remove ( completed )
        self:stopAction ( action )
    end
    self.queue.paused = false
    self.queue:_process ()
end

function app:startAction ( action )
    self._actions [ action ] = true
end

function app:stopAction ( action )
    self._actions [ action ] = nil
end

