gui =
{
    E_TOUCH = Event.declare "gui.Touch"
        :types
        {
            BEGAN   = 1,
            MOVED   = 2,
            ENDED   = 3,
            CANCEL  = 4,
        };

    E_ACTION = Event.declare "gui.Action";
    E_SCENE = Event.declare "gui.Scene"
        :types
        {
            SHOW = 1,
            HIDE = 2,
        };


    _focus = nil;
    _scenes = {};

    _sprites = {};
    -- _transform = MOAITransform2D.new ();

    textureCache = ResourceCache.new (
        function ( key )
            local texture = MOAITexture.new ()
            texture:load ( key )

            -- local filterMin = rsrc.filterMin or rsrc.filter or MOAITexture.GL_LINEAR
            -- local filterMag = rsrc.filterMag or rsrc.filter or MOAITexture.GL_LINEAR
            -- texture:setFilter ( filterMin, filterMag )
            texture:setFilter ( MOAITexture.GL_LINEAR, MOAITexture.GL_LINEAR )
            return texture
        end
    );

    fontCache = ResourceCache.new (
        function ( key )
            local font = MOAIFont.new ()
            font:loadFromBMFont ( key )
            return font
        end
    )
}

function gui:init ( viewport )
    self._viewport = viewport

    local layer = MOAILayer2D.new ()
    layer:setViewport ( viewport )
    self._layer = layer

    self._renderTable = { layer }
    MOAIRenderMgr.setRenderTable ( self._renderTable )

    self:_setupInput ()
    app.queue:subscribe ( self, self.onTouchEvent, gui.E_TOUCH )
end

function gui:loadSpriteSheet ( name, textureDir )
    local data = require ( name )
    assert ( data )

    if not textureDir then
        local parts = {}
        for s in string.gmatch ( name, "([^%.]+)" ) do
            table.insert ( parts, s )
        end
        table.remove ( parts )
        textureDir = table.concat ( parts, "/" )
    end
    local texture = textureDir .. "/" .. data.texture

    local frames = data.frames
    for _, frame in ipairs ( frames ) do
        local uv, rect = frame.uvRect, frame.spriteColorRect
        local u0, v0, u1, v1 = uv.u0, uv.v1, uv.u1, uv.v0

        if not frame.textureRotated then
            uv.quad = { u0, v1, u1, v1, u1, v0, u0, v0, }
        else
            uv.quad = { u1, v1, u1, v0, u0, v0, u0, v1, }
        end

        rect.x0, rect.y0 = rect.x, rect.y
        rect.x1, rect.y1 = rect.x + rect.width, rect.y + rect.height

        local info =
        {
            name    = frame.name,
            texture = texture,
            size    = frame.spriteSourceSize,
            uv      = uv,
            rect    = rect,
        }

        self._sprites [ frame.name ] = info
        print ( "loaded sprite frame ", frame.name )
    end
end

function gui:getSpriteInfo ( name )
    return self._sprites [ name ]
end

function gui:_setupInput ()
    local touches = { n = 0 }

    local layer = self._layer

    if MOAIInputMgr.device.pointer then
        MOAIInputMgr.device.mouseLeft:setCallback (
            function ( isMouseDown )
                local x, y = MOAIInputMgr.device.pointer:getLoc ()
                x, y = layer:wndToWorld ( x, y )
                if isMouseDown then
                    local touch = { x = x, y = y }
                    touches [ 1 ] = touch
                    touches.n = 1
                    app.queue:postEvent ( nil, gui.E_TOUCH.BEGAN, { index = 1, x = x, y = y, touches = touches })
                else
                    touches [ 1 ] = nil
                    touches.n = 0
                    app.queue:postEvent ( nil, gui.E_TOUCH.ENDED, { index = 1, x = x, y = y, touches = touches })
                end
            end
        )

        MOAIInputMgr.device.pointer:setCallback (
            function ( x, y )
                x, y = layer:wndToWorld ( x, y )
                if touches.n > 0 then
                    local touch = touches [ 1 ]
                    touch.x, touch.y = x, y
                    app.queue:postEvent ( nil, gui.E_TOUCH.MOVED, { index = 1, x = x, y = y, touches = touches })
                end
            end
        )
    else
        MOAIInputMgr.device.touch:setCallback (
            function ( eventType, idx, x, y, tapCount )
                x, y = layer:wndToWorld ( x, y )
                if eventType == MOAITouchSensor.TOUCH_DOWN then
                    local touch = { x = x, y = y }
                    touches [ idx ] = touch
                    touches.n = touches.n + 1
                    app.queue:postEvent ( nil, gui.E_TOUCH.BEGAN, { index = idx, x = x, y = y, touches = touches })
                elseif eventType == MOAITouchSensor.TOUCH_UP then
                    touches [ idx ] = nil
                    touches.n = touches.n - 1
                    app.queue:postEvent ( nil, gui.E_TOUCH.ENDED, { index = idx, x = x, y = y, touches = touches })
                elseif eventType == MOAITouchSensor.TOUCH_MOVE then
                    local touch = touches [ idx ]
                    touch.x, touch.y = x, y
                    app.queue:postEvent ( nil, gui.E_TOUCH.MOVED, { index = idx, x = x, y = y, touches = touches })
                elseif eventType == MOAITouchSensor.TOUCH_CANCEL then
                    print "touch.cancel"
                end
            end
        )
    end
end

function gui:onTouchEvent ( sender, event, args )
    local touches = args.touches
    
    if touches.n == 1 and  event == gui.E_TOUCH.BEGAN then
        -- select focused widget
        local x, y = args.x, args.y
        for i = #self._scenes, 1, -1 do
            local scene = self._scenes [ i ]
            local listeners = scene:getInputListeners ()

            for _, listener in ipairs ( listeners ) do
                if listener:acceptsTouch ( x, y ) then
                    self._focus = listener
                end
            end

            if self._focus then
                break
            end
        end
    end

    if self._focus then
        local focus = self._focus
        local handler
        if event == gui.E_TOUCH.BEGAN then
            handler = focus.onTouchBegan or focus.onTouchEvent
        elseif event == gui.E_TOUCH.MOVED then
            handler = focus.onTouchMoved or focus.onTouchEvent
        elseif event == gui.E_TOUCH.ENDED then
            handler = focus.onTouchEnded or focus.onTouchEvent
        end

        if handler then
            handler ( focus, sender, event, args )
        end
    end

    if touches.n == 0 and event == gui.E_TOUCH.ENDED then
        -- last touch was released
        self._focus = nil
    end
end

function gui:show ( scene )
    local root = scene.root
    local renderTable = self._renderTable

    table.insert ( self._scenes, scene )

    local index = #self._scenes

    assert ( not root._node.parent )
    assert ( not self._scenes [ scene ])

    self._scenes [ scene ] = index
    
    table.insert ( renderTable, index + 1, root._node.renderTable )

    -- root.transform:setParent ( self._transform )
end

function gui:hide ( scene )
    local root = scene.root
    local renderTable = self._renderTable

    -- root.transform:setParent ( nil )

    local index = self._scenes [ scene ]
    assert ( index )

    table.remove ( self._scenes, index )
    table.remove ( renderTable, index + 1 )

    self._scenes [ scene ] = nil
    for i = index, #self._scenes do
        local s = self._scenes [ i ]
        self._scenes [ s ] = i
    end
end


gui.Widget = core.class ( Object )
{
    PROP_FACTORY = MOAIProp2D;
}

function gui.Widget:ctor ( rsrc, context )
    assert ( rsrc )
    context = context or { root = self }

    self:_processID ( rsrc [ rm.ID ], context )
    self.rsrc = rsrc

    self._node = { renderTable = {}, renderIndex = nil }

    if rsrc.deck or self.PROP_FACTORY ~= MOAIProp2D then
        local prop = self.PROP_FACTORY.new ()
        if rsrc.deck then
            local deck = rsrc.deck:create ()
            prop:setDeck ( deck )
        end
		
        self.prop       = prop
        self.transform  = prop
		
		if rsrc.scissor then
			local scissorRect = MOAIScissorRect.new ()
			scissorRect:setRect ( unpack ( rsrc.scissor ) )
			self.prop:setScissorRect( scissorRect )
		end
		
		if rsrc.blend then
			local src = rsrc.blend.src or MOAIProp2D.GL_ONE
			local dst = rsrc.blend.dst or MOAIProp2D.GL_ONE_MINUS_SRC_ALPHA
			self.prop:setBlendMode ( src, dst )
		end
		
        table.insert ( self._node.renderTable, prop )
    else
        self.transform  = MOAITransform2D.new ()
    end
	
	if rsrc.grid then
		local grid = rsrc.grid:create ()
		self.prop:setGrid ( grid )
	end	
	
	if rsrc.shader then
		self.prop:setShader ( rsrc.shader )
	end
	
    -- set location
    local x, y = rsrc.x or 0, rsrc.y or 0
    local loc = rsrc.xy or rsrc.loc
    if loc then
        x, y = loc [ 1 ], loc [ 2 ]
    end
    self.transform:setLoc ( x, y )

    local piv = rsrc.piv or { 0, 0 }
    self.transform:setPiv ( piv[ 1 ], piv [ 2 ] )

    local scaleX = rsrc.scaleX or rsrc.scale or 1
    local scaleY = rsrc.scaleY or rsrc.scale or 1
    self.transform:setScl ( scaleX, scaleY )

    local rot = rsrc.rot or rsrc.rotation or 0
    self.transform:setRot ( rot )

    if rsrc.index then
        self.prop:setIndex ( rsrc.index )
    end

    if rsrc.color and self.prop then
        self.prop:setColor ( unpack ( rsrc.color ) )
    end

	if rsrc.priority and self.prop then
        self.prop:setPriority ( rsrc.priority )
    end
	
    local last = context.last
    context.last = self
    for _, r in ipairs ( rsrc ) do
        local child = r:create ( context )
        child:attach ( self )
    end
    context.last = last
end

function gui.Widget:_processID ( id, context )
    if not id then
        return
    end
    local start = 1
    local lo, hi
    local target, key
    if string.find ( id, [[^/]] ) then
        target = context.root
        start = 2
    else
        target = context.last
        lo, hi = string.find ( id, [[^%.+]] )
        if lo and hi then
            for i = 1, math.max ( hi - lo, 0 ) do
                assert ( target )
                target = target._node.parent
            end
            start = hi + 1
        end
    end

    while true do
        lo = string.find ( id, [[/]], start )
        if not lo then
            break
        end

        start = lo + 1
        local k = string.sub ( id, start, lo - 1 )
        target = target [ k ]
        assert ( target )
    end
    key = string.sub ( id, start )

    if target and key then
        target [ key ] = self
    end
end

function gui.Widget:attach ( parent, before )
    assert ( parent and self )
    if before then
        assert ( before._node.parent == parent )
    end

    local node0 = self._node
    local nodeP = parent._node

    local rght = before
    local left = ( before and before._node.leftBrother ) or ( not before and nodeP.lastChild )

    local nodeR = rght and rght._node
    local nodeL = left and left._node

    node0.parent = parent
    node0.leftBrother = left
    node0.rghtBrother = rght

    if left then
        left._node.rghtBrother = self
    end
    if rght then
        rght._node.leftBrother = self
    end

    if left == nodeP.lastChild then
        nodeP.lastChild = self
    end

    if rght == nodeP.frstChild then
        nodeP.frstChild = self
    end

    if rght then
        node0.renderIndex = rght._node.renderIndex
        table.insert ( nodeP.renderTable, node0.renderIndex, node0.renderTable )
        local wdg = rght
        while wdg do
            local node = wdg._node
            node.renderIndex = node.renderIndex + 1
            wdg = node.rghtBrother
        end
    else
        table.insert ( nodeP.renderTable, node0.renderTable )
        node0.renderIndex = #nodeP.renderTable
    end

    self.transform:setParent ( parent.transform )
    self.transform:forceUpdate ()
    return self
end

function gui.Widget:detach ()
    local node0 = self._node
    
    if not node0.parent then
        return self
    end

    local prnt, left, rght = node0.parent, node0.leftBrother, node0.rghtBrother
    local nodeP = prnt._node

    if left then
        left._node.rghtBrother = rght
    end

    if rght then
        rght._node.leftBrother = left
    end

    node0.rghtBrother   = nil
    node0.leftBrother   = nil
    node0.parent        = nil

    if nodeP.frstChild == self then
        nodeP.frstChild = rght
    end

    if nodeP.lastChild == self then
        nodeP.lastChild = left
    end
    
    table.remove ( nodeP.renderTable, node0.renderIndex )
    node0.renderIndex = nil
    
    while rght do
        local node = rght._node
        node.renderIndex = node.renderIndex - 1
        rght = node.rghtBrother
    end

    self.transform:setParent ( nil )
    self.transform:forceUpdate ()
    return self
end

function gui.Widget:getParent ()
    return self._node.parent
end

local function children ( root, wdg )
    if not wdg then
        return root._node.frstChild
    end
    wdg = wdg._node.rghtBrother
    return wdg
end

local function preorder ( root, wdg )
    if not wdg then
        return root
    end

    local child = wdg._node.frstChild

    if child then
        return child
    end

    local brother = wdg._node.rghtBrother
    while not brother do
        if wdg == root then
            return nil
        end
        wdg = wdg._node.parent
        if not wdg then
            return nil
        end
        brother = wdg._node.rghtBrother
    end
    return brother
end

function gui.Widget:iterChildren ()
    return children, self, nil
end

function gui.Widget:walk ()
    return preorder, self, nil
end

function gui.Widget:getChildren ()
    local ans = {}
    for child in self:iterChildren () do
        table.insert ( ans, child )
    end
    return ans
end

function gui.Widget:acceptsTouch ( x, y )
    return false
end


gui.Button = core.class ( gui.Widget )
{
    REGULAR = 1,
    PRESSED = 2,
    DISABLED = 3;
}

function gui.Button:ctor ( rsrc, context )
    gui.Widget.ctor ( self, rsrc, context )

    assert ( rsrc.id )
    self.id = rsrc.id

    self._isPressed = false
    self._isEnabled = true
    self:_update ( )
end

function gui.Button:_update ()
    local state
    if self._isEnabled then
        state = self._isPressed and self.PRESSED or self.REGULAR
    else
        state = self.DISABLED
    end

    if self._state == state then
        return
    end

    self._state = state
    self:updateLook ()
end

function gui.Button:updateLook ()
    local state = self._state
    if state then
        self.prop:setIndex ( state )
    end
end

function gui.Button:acceptsTouch ( x, y )
    if not self:getParent () then
        return false
    end
    if not self._isEnabled then
        return false
    end
    return self.prop:inside ( x, y )
end

function gui.Button:onTouchBegan ( sender, event, args )
    local touches = args.touches
    if touches.n > 1 then
        return
    end
    self._touchIndex = args.index
    self._isPressed = true
    self:_update ()
end

function gui.Button:onTouchMoved ( sender, event, args )
    if args.index ~= self._touchIndex then
        return
    end

    self._isPressed = self.prop:inside ( args.x, args.y )
    self:_update ()
end

function gui.Button:onTouchEnded ( sender, event, args )
    if args.index ~= self._touchIndex then
        return
    end

    if self._isPressed then
        self._isPressed = false
        self:_update ()
        app.queue:postEvent ( self, gui.E_ACTION, { id = self.id })
    end    
end

function gui.Button:setEnabled ( enabled )
    if not enabled == not self._isEnabled then
        return
    end
    if not enabled then
        self._touchIndex = nil
        self._isPressed = nil
    end
    self._isEnabled = enabled
    self:_update ()
end


gui.TextBox = core.class ( gui.Widget )
{
    PROP_FACTORY = MOAITextBox;
}

function gui.TextBox:ctor ( rsrc, context )
    gui.Widget.ctor ( self, rsrc, context )

    local fontSize = rsrc.fontSize
    local font = gui.BMFont.new { src = rsrc.font }
    
    local style = MOAITextStyle.new ()
    style:setFont ( font )
    style:setSize ( fontSize )
    
    if rsrc.color then
        style:setColor(unpack ( rsrc.color ))
    end
    
    self.prop:setStyle ( style )
    
    local textAlignX = rsrc.textAlignX or MOAITextBox.CENTER_JUSTIFY
    local textAlignY = rsrc.textAlignY or MOAITextBox.CENTER_JUSTIFY
    self.prop:setAlignment ( textAlignX, textAlignY )
    
    local flipY = true
    if rsrc.flipY ~= null then
        flipY = rsrc.flipY
    end
    self.prop:setYFlip ( flipY )  
    
    if rsrc.rect then
        self.prop:setRect ( unpack ( rsrc.rect ))
    end
    
    if rsrc.text then
        self.prop:setString ( rsrc.text )
    end
end


gui.Scene = core.class ( Object )
{
    STATE_NONE = 1,
    STATE_SHOW = 2,
    STATE_TRANSITION_SHOW = 3,
    STATE_TRANSITION_HIDE = 4;
}

function gui.Scene:ctor ( root )
    self._state = self.STATE_NONE
    self._inputListeners = { self }
    self.root = root
end

function gui.Scene:show ( zIndex )
    if self._state ~= self.STATE_NONE then
        return
    end
    self._state = self.STATE_TRANSITION_SHOW
    gui:show ( self )
    self:onShowTransition ()
end

function gui.Scene:onShowTransition ()
    self:onShow ()
end

function gui.Scene:onShow ()
    assert ( self._state == self.STATE_TRANSITION_SHOW )
    app.queue:postEvent ( self, gui.E_SCENE.SHOW, {} )
    self._state = self.STATE_SHOW
end

function gui.Scene:hide ( callback )
    if self._state ~= self.STATE_SHOW then
        return
    end
    self._cbHide = callback
    self._state = self.STATE_TRANSITION_HIDE
    self:onHideTransition ()
end

function gui.Scene:onHideTransition ()
    self:onHide ()
end

function gui.Scene:onHide ()
    assert ( self._state == self.STATE_TRANSITION_HIDE )
    self._state = self.STATE_NONE
    gui:hide ( self )
    app.queue:postEvent ( self, gui.E_SCENE.HIDE, {} )
    if self._cbHide then
        self._cbHide ()
        self._cbHide = nil
    end
end

function gui.Scene:acceptsTouch ( x, y )
    return true
end

function gui.Scene:getInputListeners ()
    return self._inputListeners
end


-- factories
gui.Texture =
{
    new = function ( rsrc )
        return gui.textureCache:get ( rsrc.name )
    end;
}

gui.VideoTexture =
{
    new = function ( rsrc )
		local texture = MOAIVideoTexture.new ()
		
		if rsrc.type then
			texture:setType ( rsrc.type ) 
		end
		
        return texture
    end;	
}

gui.MultiTexture =
{
    new = function ( rsrc )
        local multiTexture = MOAIMultiTexture.new ()
		
        local cnt = #rsrc
        assert ( cnt > 0 )
        multiTexture:reserve ( cnt )
		
        for i, v in ipairs ( rsrc ) do
			local texture
			
			-- use the entire texture
			if type ( v ) == "string" then
				texture = gui.Texture.new { name = v }
			else
				texture = v:create ()
			end
			
			assert ( texture )			
			multiTexture:setTexture ( i, texture )
        end

        return multiTexture
    end;
}

gui.BMFont =
{
    new = function ( rsrc )
        return gui.fontCache:get ( rsrc.src )
    end;
}


gui.GfxQuad =
{
    new = function ( rsrc )
        local deck = MOAIGfxQuad2D.new ()

        local texture, rect

        if rsrc.sprite then
            local info = gui:getSpriteInfo ( rsrc.sprite )

            texture = gui.Texture.new { name = info.texture }

            deck:setTexture ( texture )
            deck:setUVQuad ( unpack ( info.uv.quad ))

            local w, h = 0.5 * info.size.width, 0.5 * info.size.height
            local x, y = rsrc.x or 0, rsrc.y or 0
            local alignX, alignY = rsrc.alignX or 0, rsrc.alignY or 0
            local scaleX, scaleY = rsrc.scaleX or 1, rsrc.scaleY or 1

            local w_, h_ = w * scaleX, h * scaleY
            local x_ = x - alignX * w_
            local y_ = y - alignY * h_

            local rect = info.rect
            local x0, y0, x1, y1 = rect.x0, rect.y0, rect.x1, rect.y1

            x0 = x_ + ( x0 - w ) * scaleX
            x1 = x_ + ( x1 - w ) * scaleX
            y0 = y_ + ( y0 - h ) * scaleY
            y1 = y_ + ( y1 - h ) * scaleY

            deck:setRect ( x0, y0, x1, y1 )
        else
            -- use the entire texture
            if type ( rsrc.texture ) == "string" then
                texture = gui.Texture.new { name = rsrc.texture }
            else
                texture = rsrc.texture:create ()
            end

            -- texture
            deck:setTexture ( texture )
			
			if rsrc.uvRect then
				deck:setUVRect ( unpack ( rsrc.uvRect ) )
			else
				deck:setUVRect ( 0, 1, 1, 0 )
			end

			-- drawing rect
			if rsrc.rect then
				deck:setRect ( unpack ( rsrc.rect ) )
			else
				local w, h = texture:getSize ()
				w, h = 0.5 * w, 0.5 * h
				local x, y = rsrc.x or 0, rsrc.y or 0
				local alignX, alignY = rsrc.alignX or 0, rsrc.alignY or 0
				local scaleX, scaleY = rsrc.scaleX or 1, rsrc.scaleY or 1

				local w_, h_ = w * scaleX, h * scaleY
				local x_ = x - alignX * w_
				local y_ = y - alignY * h_

				local x0, y0, x1, y1 = 0, 0, 2 * w, 2 * h

				x0 = x_ + ( x0 - w ) * scaleX
				x1 = x_ + ( x1 - w ) * scaleX
				y0 = y_ + ( y0 - h ) * scaleY
				y1 = y_ + ( y1 - h ) * scaleY

				deck:setRect ( x0, y0, x1, y1 )
			end
        end

        return deck
    end;
}


gui.GfxQuadDeck =
{
    new = function ( rsrc )
        local deck = MOAIGfxQuadDeck2D.new ()
        local texture

        local cnt = #rsrc
        assert ( cnt > 0 )
        deck:reserve ( cnt )

        for i, quad in ipairs ( rsrc ) do
            local info
            if type ( quad ) == "string" then
                info = gui:getSpriteInfo ( quad )
                quad = {}
            else
                assert ( quad.sprite )
                info = gui:getSpriteInfo ( quad.sprite )
            end

            if not texture then
                texture = gui.Texture.new { name = info.texture }
            else
                assert ( gui.textureCache:get ( info.texture ) == texture, string.format ( "Sprite %s is in wrong spritesheet", info.name ))
            end

            deck:setUVQuad ( i, unpack ( info.uv.quad ))

            local w, h = 0.5 * info.size.width, 0.5 * info.size.height
            local x, y = quad.x or rsrc.x or 0, quad.y or rsrc.y or 0
            local alignX, alignY = quad.alignX or rsrc.alignX or 0, quad.alignY or rsrc.alignY or 0
            local scaleX, scaleY = quad.scaleX or rsrc.scaleX or 1, quad.scaleY or rsrc.scaleY or 1

            local w_, h_ = w * scaleX, h * scaleY
            local x_ = x - alignX * w_
            local y_ = y - alignY * h_

            local rect = info.rect
            local x0, y0, x1, y1 = rect.x0, rect.y0, rect.x1, rect.y1

            x0 = x_ + ( x0 - w ) * scaleX
            x1 = x_ + ( x1 - w ) * scaleX
            y0 = y_ + ( y0 - h ) * scaleY
            y1 = y_ + ( y1 - h ) * scaleY

            deck:setRect ( i, x0, y0, x1, y1 )
        end

        assert ( texture )
        deck:setTexture ( texture )

        return deck
    end;
}

gui.GfxQuadListDeck =
{
    new = function ( rsrc )
        local deck = MOAIGfxQuadListDeck2D.new ()
        local texture

        local listsCnt = 0
        local pairsCnt = 0
        local maxIndex = 0
        local xyRects, uvQuads = {}, {}

        for i, list in ipairs ( rsrc ) do
            if type( i ) == "number" then
                assert ( i > 0 )

                listsCnt, pairsCnt = listsCnt + 1, pairsCnt + #list

                for iq, quad in ipairs ( list ) do
                    local sprite = quad.sprite
                    assert ( sprite )

                    local info = gui:getSpriteInfo ( sprite )
                    if not texture then
                        texture = gui.Texture.new { name = info.texture }
                    else
                        assert ( gui.textureCache:get ( info.texture ) == texture )
                    end

                    if not uvQuads [ sprite ] then
                        table.insert ( uvQuads, info.uv.quad )
                        uvQuads [ sprite ] = #uvQuads
                    end

                    local w, h = 0.5 * info.size.width, 0.5 * info.size.height
                    local x, y = quad.x or rsrc.x or 0, quad.y or rsrc.y or 0
                    local alignX, alignY = quad.alignX or rsrc.alignX or 0, quad.alignY or rsrc.alignY or 0
                    local scaleX, scaleY = quad.scaleX or rsrc.scaleX or 1, quad.scaleY or rsrc.scaleY or 1

                    local w_, h_ = w * scaleX, h * scaleY
                    local x_ = x - alignX * w_
                    local y_ = y - alignY * h_

                    local rect = info.rect
                    local x0, y0, x1, y1 = rect.x0, rect.y0, rect.x1, rect.y1

                    x0 = x_ + ( x0 - w ) * scaleX
                    x1 = x_ + ( x1 - w ) * scaleX
                    y0 = y_ + ( y0 - h ) * scaleY
                    y1 = y_ + ( y1 - h ) * scaleY

                    table.insert ( xyRects, { x0, y0, x1, y1 })
                    xyRects [ quad ] = #xyRects
                end
            end
        end

        assert ( texture )
        deck:setTexture ( texture )

        deck:reserveLists ( listsCnt )
        deck:reservePairs ( pairsCnt )
        deck:reserveQuads   ( #xyRects )
        deck:reserveUVQuads ( #uvQuads )

        for _, rect in ipairs ( xyRects ) do
            deck:setRect ( _, unpack ( rect ))
        end

        for _, quad in ipairs ( uvQuads ) do
            deck:setUVQuad ( _, unpack ( quad ))
        end

        local base, rectId = 1, 1
        for i, list in ipairs ( rsrc ) do
            deck:setList ( i, base, #list )
            for iq, quad in ipairs ( list ) do
                deck:setPair ( base + iq - 1, uvQuads [ quad.sprite ], xyRects [ quad ])
            end
            base = base + #list
        end

        return deck
    end;
}

gui.TileDeck = 
{
    new = function ( rsrc )
        local deck = MOAITileDeck2D.new ()

        local texture, rect
        local u0, v0, u1, v1
			
		if type ( rsrc.texture ) == "string" then
			texture = gui.Texture.new { name = rsrc.texture }
		else
			texture = rsrc.texture:create ()
		end

		deck:setTexture ( texture )

		-- UV rect
		-- u0 = rsrc.u0 or 0
		-- v0 = rsrc.v0 or 1
		-- u1 = rsrc.u1 or 1
		-- v1 = rsrc.v1 or 0   
		-- if rsrc.uv then
		--     u0, v0, u1, v1 = unpack ( rsrc.uv )
		-- end         
		--deck:setUVRect ( u0, v0, u1, v1 )

		-- tiles count
		local rows = rsrc.rows or 1
		local cols = rsrc.cols or 1
		deck:setSize ( cols, rows )		
		
		-- -- drawing rect
		-- if rsrc.rect then
			-- deck:setRect ( unpack ( rsrc.rect ))
		-- else
			-- local w, h = texture:getSize ()
			-- w, h = w / cols, h / rows

			-- local alignX = rsrc.alignX or rsrc.align or 0
			-- local alignY = rsrc.alignY or rsrc.align or 0

			-- local x0, x1 = 0.5 * w * ( -1 - alignX ), 0.5 * w * ( 1 - alignX )
			-- local y0, y1 = 0.5 * h * ( -1 - alignY ), 0.5 * h * ( 1 - alignY )
			-- deck:setRect ( x0, y0, x1, y1 )
		-- end
		
        return deck
    end;
}

gui.RectGrid = 
{
    new = function ( rsrc )
        local grid = MOAIGrid.new ()

        local width = rsrc.width or 1
        local height = rsrc.height or 1		
		
		local tileWidth = rsrc.tileWidth or 1
		local tileHeight = rsrc.tileHeight or 1
		
		grid:initRectGrid ( width, height, tileWidth, tileHeight )
		
		if rsrc.fill then
			grid:fill ( rsrc.fill )
		end

        return grid
    end;	
}

-- action factories
local function createAction ( rc, target, options )
    if type ( rc ) == "function" then
        local a = rc ( target, options )
        a:stop ()
        return a
    end
    return rc:create ( target, options )
end


gui.ChainAction =
{
    new = function ( cfg, target, options )
        local action = MOAIAction.new ()
    
        local children = {}
        for _, rc in ipairs ( cfg ) do
            local a = createAction ( rc, target, options )
            table.insert ( children, a )
        end
    
        for i = 1, #children do
            local currAction = children [ i ]
            local nextAction = children [ i + 1 ]
            if i == 1 then
                action:addChild ( currAction )
            end

            if nextAction then
                currAction:setListener ( MOAIAction.EVENT_STOP,
                    function ()
                        action:addChild ( nextAction )
                    end
                )
            elseif cfg.onDone then
                currAction:setListener ( MOAIAction.EVENT_STOP,
                    function ()
                        cfg.onDone ( target, options )
                    end
                )
            end
        end
        
        return action
  end;
}


gui.SplitAction =
{
    new = function ( cfg, target, options )
        local action = MOAIAction.new ()
        for _, rc in ipairs ( cfg ) do
            local a = createAction ( rc, target, options )
            action:addChild ( a )
        end
        return action
    end;
}


gui.WaitAction =
{
    new = function ( cfg, target, options )
        local timer = MOAITimer.new ()
        local t = type ( cfg.t )
        local dt
        if t == "table" then
            dt = cfg.t
        elseif t == "function" then
            dt = cfg.t ( target, options )
        elseif t == "number" then
            dt = cfg.t
        end
        timer:setSpan ( dt )
        return timer
    end
}


gui.TimelineAction =
{
    new = function ( cfg, target, options )
        local count = #cfg
        assert ( count > 0 )
        
        local duration = cfg.duration
        local loopsCnt = cfg.loopsCnt or 1

        local currIdx = #cfg + 1
        local currLoop = 0
        local timer = MOAITimer.new ()
      
        local function onTimeout ()
            local currNode = cfg [ currIdx ]

            if not currNode then
                -- start next loop
                if currLoop == loopsCnt and loopsCnt > 0 then
                -- done
                    timer:stop ()
                    return
                end

                local dt = math.max ( cfg[ 1 ].t, 0.01 )
                timer:setSpan ( dt )
                currLoop, currIdx = currLoop + 1, 1
                return
            end

            local t, func = currNode.t, currNode [ 1 ]
            func ( target, options )
            currIdx = currIdx + 1
            if currIdx <= count then
                local dt = cfg [ currIdx ].t - t
                timer:setSpan ( dt )
            else
                if not duration then
                    timer:stop ()
                else
                    -- wait for loop to end
                    timer:setSpan ( duration - currNode.t )
                end
            end
        end

        timer:setMode ( MOAITimer.LOOP )
        timer:setListener ( MOAITimer.EVENT_TIMER_END_SPAN, onTimeout)
        
        -- start first loop
        onTimeout ()
        
        return timer
    end
}

