core = {}

unpack = unpack or table.unpack

function log ( fmt, ... )
    local msg = string.format ( fmt, ... )
    print ( msg )
end

function core.filter ( t, predicate, inplace )
    local n = #t
    if inplace then
        for i = n, 1, -1 do
            if not predicate ( t [ i ] ) then
                table.remove ( t, i )
            end
        end
        return t
    else
        local filtered = {}
        for i = 1, n do
            local v = t [ i ]
            if predicate ( v ) then
                table.insert ( filtered, v )
            end
        end
        return filtered
    end
end

function deepcopy(t)
    if type(t) ~= 'table' then return t end
    local mt = getmetatable(t)
    local res = {}
    for k,v in pairs(t) do
        if type(v) == 'table' then
            v = deepcopy(v)
        end
        res[k] = v
    end
    setmetatable(res,mt)
    return res
end

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- classes
------------------------------------------------------------------------------------
core.class = function ( base )
    local imembers = base and setmetatable ({}, { __index = base.__imembers__ }) or {}
    local imt = { __index = imembers }

    local C = { __imembers__ = imembers }

    local function override ( cls, k, v )
        rawset ( cls, k, v )
        imembers [ k ] = v
    end

    local function statics ( cls, members )
        for k, v in pairs ( members ) do
            override ( C, k, v )
        end

        return C
    end

    local cmt =
    {
        __index = base,
        __newindex = override,
        __call = statics,
    }

    function C.alloc ( o )
        return setmetatable ( o or {}, imt )
    end

    function C.new ( ... )
        local o = C.alloc ()
        o:ctor ( ... )
        return o
    end

    return setmetatable ( C, cmt )
end


Object = core.class ()

function Object:ctor ( self )
end

function Object:toString ( self )
    return tostring ( self )
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- observer pattern (synchronous callbacks)
------------------------------------------------------------------------------------
local function connect ( self, observer, method )
    local t = { o = observer, m = method }
    table.insert ( self, t )
end

local function disconnect ( self, observer, method )
    local n = #self
    for i = n, 1, -1 do
        if ( not observer or self [ i ].o == observer ) and ( not method or self [ i ].m == method ) then
            table.remove ( self, i )
        end
    end
end

local mt =
{
    __call = function ( self, ... )
        for _, callback in ipairs ( self ) do
            callback.m ( callback.o, ... )
        end
    end
}

local mt0 =
{
    __call = function ( self )
        for _, callback in ipairs ( self ) do
            callback.m ( callback.o )
        end
    end
}

local mt1 =
{
    __call = function ( self, arg1 )
        for _, callback in ipairs ( self ) do
            callback.m ( callback.o, arg1 )
        end
    end
}

function core.signal ()
    return setmetatable ({ connect = connect, disconnect = disconnect } , mt )
end

function core.signal0 ()
    return setmetatable ({ connect = connect, disconnect = disconnect }, mt0 )
end

function core.signal1 ()
    return setmetatable ({ connect = connect, disconnect = disconnect }, mt1 )
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- events
------------------------------------------------------------------------------------
Event = core.class ( Object )
{
    declare = function ( name )
        local e = Event.new ()

        e.class = e
        e._description = name or e

        e.types = function ( self, types )
            for k, v in pairs ( types ) do
                local subevent = Event.new ()
                subevent.class = e
                subevent._description = string.format ( "%s:%s", e._description, k )
                self [ k ] = subevent
            end
            return self
        end
        return e
    end;
}

function Event:toString ()
    return self._description
end


EventQueue = core.class ( Object )

function EventQueue:ctor ()
    self.processed = core.signal0 ()
    self._head = 0
    self._tail = 0
    self._byEvent   = {}
    self._bySender  = {}
    self.paused = false
end

function EventQueue:subscribe ( object, method, event, sender )
    assert ( object and method )
    assert ( event or sender )
    local handler = { o = object, m = method, s = sender, e = event }
    if event then
        local handlers = self._byEvent [ event.class ] or {}
        table.insert ( handlers, handler )
        self._byEvent [ event.class ] = handlers
    end
    if sender then
        local handlers = self._bySender [ sender ] or {}
        table.insert ( handlers, handler )
        self._bySender [ sender ] = handlers
    end
end

function EventQueue:unsubscribe ( object, method, event, sender )
    assert ( object )
    local handlers
    handlers = event and self._byEvent [ event.class ] or nil
    if handlers then
        core.filter ( handlers, function ( h )
            if h.o ~= object then return true end
            if not method then return false end
            return h.m ~= method
        end, true )
    end
    handlers = sender and self._bySender [ sender ] or nil
    if handlers then
        core.filter ( handlers, function ( h )
            if h.o ~= object then return true end
            if not sender then return false end
            return h.s ~= sender
        end, true )
    end
end

function EventQueue:postEvent ( sender, event, args )
    assert ( event )
    args = args or {}
    local head, tail = self._head, self._tail
    local msg = { s = sender, e = event, d = args }
    local cnt = tail - head

    self [ tail ] = msg
    self._tail = tail + 1

    if cnt == 0 and not self.paused then
        self:_process ()
    end
end

function EventQueue:_process ()
    local function invoke (h, s, e, d)
        if h.s and h.s ~= s then return nil end
        if h.e and h.e ~= e and h.e ~= e.class then return nil end
        h.m ( h.o, s, e, d )
        return true
    end

    local done = self._head == self._tail
    local handlers, msg

    while not done do
        msg = self [ self._head ]

        local invoked = {}
        handlers = self._byEvent [ msg.e.class ]
        if handlers then
            for _, h in ipairs ( handlers ) do
                invoked [ h ] = invoked [ h ] or invoke ( h, msg.s, msg.e, msg.d )
            end
        end
        handlers = msg.s and self._bySender [ msg.s ] or nil
        if handlers then
            for _, h in ipairs ( handlers ) do
                invoked [ h ] = invoked [ h ] or invoke ( h, msg.s, msg.e, msg.d )
            end
        end

        -- remove message from the queue only after processing
        self._head = self._head + 1
        done = self._head == self._tail
    end
    self._head, self._tail = 0, 0
    self.processed ()
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- State machine
------------------------------------------------------------------------------------
FSM = core.class ( Object )

function FSM:ctor ( context )
    self._context = context
    self._locked = false
end

function FSM:declare ( states )
    self._states = states
    self._currState = states [ 1 ] or states.initial
    assert ( self._currState, "No initial state found" )
end

function FSM:_exec ( state, sender, event, args )
    assert ( event )
    assert ( not self._locked )
    self._locked = true

    local handler = state [ event ] or state [ event.class ]

    local result
    if handler then
        result = handler ( self._context, sender, event, args )
        if type ( result ) == "string" then
            local stateName = result
            result = self._states [ stateName ]
            assert ( result, "state not found: " .. tostring( stateName ))
        end
    end
    self._locked = false
    return result
end

function FSM:handleEvent ( sender, event, args )
    local currState = self._currState
    assert ( event )
    assert ( currState )
    local nextState = self:_exec ( currState, sender, event, args )
    while nextState do
        self:_exec ( currState, self, "leave" )
        currState = nextState
        nextState = self:_exec ( currState, self, "enter" )
    end
    self._currState = currState
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- Resource manager
------------------------------------------------------------------------------------
rm =
{
    ID = "@ID";
}

-- declare resource
function rm:rsrc ( factory, id )
    assert ( factory )
    local lookup, rm = self._lookup, self
    return function ( cfg )
        local rsrc = cfg
        rsrc [ rm.ID ] = id
        rsrc.create = function ( self, ... )
            return factory.new ( self, ... )
        end
        return rsrc
    end
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- resource cache
------------------------------------------------------------------------------------
ResourceCache = core.class ( Object )

function ResourceCache:ctor ( creator )
    self._lookup = setmetatable( {}, { __mode = 'v' })
    self._creator = creator
end

function ResourceCache:get ( key )
    local resource = self._lookup [ key ]
    if not resource then
        resource = self._creator ( key )
        self._lookup [ key ] = resource
    end
    return resource
end

function ResourceCache:set ( key, resource )
    self._lookup [ key ] = resource
end
------------------------------------------------------------------------------------


------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
-- application object
------------------------------------------------------------------------------------
E_SYSTEM = Event.declare "system"
    :types
    {
        START = 1,
        QUIT = 2,
        PAUSE = 3,
        RESUME = 4,
    }

E_TIMEOUT = Event.declare "timeout"

app =
{
    queue = EventQueue.new ();
}


Timer = core.class ( Object )

function Timer:ctor ( timeout, callback )
    assert ( timeout )
    self._timeout = timeout
    self._callback = callback
end

function Timer:start ( loopsCnt )
    local timer = MOAITimer.new ()
    local loop = 0

    self._timer = timer

    timer:setMode ( MOAITimer.LOOP )
    timer:setSpan ( self._timeout )
    timer:setListener ( MOAITimer.EVENT_TIMER_END_SPAN,
        function ()
            loop = loop + 1
            if loop == loopsCnt then
                self:stop ()
            end
            
            if self._callback then
                self._callback ( loop )
            else
                app.queue:postEvent ( self, E_TIMEOUT, { loop = loop })
            end
        end
    )
    timer:start ()
    return self
end

function Timer:stop ()
    if self._timer then
        self._timer:stop ()
    end
    return self
end

------------------------------------------------------------------------------------


NetChannel = core.class ( Object )

function NetChannel:ctor ()
    self.opened         = core.signal ()
    self.closed         = core.signal ()
    self.msgReceived    = core.signal ()
    self.stateChanged   = core.signal ()

    self._thread = MOAICoroutine.new ()
    self._state = nil

    self._buffer = {}
end

function NetChannel:close ()
    if not self._thread:isActive () then
        self._thread:stop ()
    end
end

function NetChannel:open ()
    if not MOAIP2PChannel then
        return false
    end

    assert ( not self._thread:isActive ())

    local function parse ( data )
        local msgType, msgArgs 
        local i = string.find ( data, "\n" )
        if i then
            msgType = string.sub ( data, 1, i - 1 )
            msgArgs = MOAIJsonParser.decode ( string.sub (  data, i + 1 ))
        end
        return msgType, msgArgs
    end

    self._thread:run (
        function ()
            local done = false
            local prevState, currState

            while not done do
                coroutine.yield ()

                currState = MOAIP2PChannel.getState ()
                self._state = currState

                if currState ~= prevState then
                    self.stateChanged ( prevState, currState )

                    if currState == MOAIP2PChannel.CONNECTED then
                        local peerId = MOAIP2PChannel.getPeerId ()
                        print ( prevState, currState )
                        self.opened ( peerId )
                    else
                        self.closed ()
                    end
                    prevState = currState
                end

                if currState == MOAIP2PChannel.CONNECTED then
                    local data = MOAIP2PChannel.receive ()
                    if data then
                        print ( "--- RECV" )
                        print ( data )
                        local msgType, msgArgs = parse ( data )
                        self.msgReceived ( msgType, msgArgs )
                    end
                end
            end
        end
    )
    return true
end

function NetChannel:isConnected ()
    return MOAIP2PChannel and self._state == MOAIP2PChannel.CONNECTED
end

function NetChannel:send ( msgType, msgArgs )
    if not self:isConnected () or not MOAIP2PChannel then
        return nil
    end

    local msg = string.format ( "%s\n%s", msgType, MOAIJsonParser.encode ( msgArgs ))
    print ( "--- SEND" )
    print ( msg )
    MOAIP2PChannel.send ( msg )
    return msg
end

do
    -- Set application onPause/onResume
    if MOAISim.setPauseCallback then
        MOAISim.setPauseCallback ( function ()
                                       app.queue:postEvent ( self, E_SYSTEM.PAUSE )
                                   end )
    end
    if MOAISim.setResumeCallback then
        MOAISim.setResumeCallback ( function ()
                                        app.queue:postEvent ( self, E_SYSTEM.RESUME )
                                    end )
    end
end
